#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from binascii import unhexlify, hexlify
import pytest

from .common import SafeWISETest

from safewiselib import coins
from safewiselib import messages as proto
from safewiselib.tools import parse_path

TxApiZcashTestnet = coins.tx_api['Zcash Testnet']

TXHASH_aaf51e = unhexlify('aaf51e4606c264e47e5c42c958fe4cf1539c5172684721e38e69f4ef634d75dc')


@pytest.mark.zcash
class TestMsgSigntxZcash(SafeWISETest):

    def test_one_one_fee(self):
        self.setup_mnemonic_allallall()

        # tx: aaf51e4606c264e47e5c42c958fe4cf1539c5172684721e38e69f4ef634d75dc
        # input 1: 3.0 TAZ

        inp1 = proto.TxInputType(
            address_n=parse_path("m/Zcash Testnet/0h/0/0"),  # tmQoJ3PTXgQLaRRZZYT6xk8XtjRbr2kCqwu
            amount=300000000,
            prev_hash=TXHASH_aaf51e,
            prev_index=1,
        )

        out1 = proto.TxOutputType(
            address='tmJ1xYxP8XNTtCoDgvdmQPSrxh5qZJgy65Z',
            amount=300000000 - 1940,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        with self.client:
            self.client.set_tx_api(TxApiZcashTestnet)
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])

            (signatures, serialized_tx) = self.client.sign_tx('Zcash Testnet', [inp1, ], [out1, ], version=3, overwintered=True)

        # Accepted by network: tx eda9b772c47f0c29310759960e0081c98707aa67a0a2738bcc71439fcf360675
        assert hexlify(serialized_tx) == b'030000807082c40301dc754d63eff4698ee321476872519c53f14cfe58c9425c7ee464c206461ef5aa010000006a47304402207e45f303b4e42be824513855eb21653e1d2749cd94dcd0f0613d3f85d4efd1e20220699ffbdbcad889af7ede5ce9febf7a5ef8f5619b2464824529974c400cffaebc0121030e669acac1f280d1ddf441cd2ba5e97417bf2689e4bbec86df4f831bf9f7ffd0ffffffff016c9be111000000001976a9145b157a678a10021243307e4bb58f36375aa80e1088ac000000000000000000'



# Automatically generated by pb2py
from .. import protobuf as p
from .HDNodeType import HDNodeType


class DebugLinkState(p.MessageType):
    MESSAGE_WIRE_TYPE = 102
    FIELDS = {
        1: ('layout', p.BytesType, 0),
        2: ('pin', p.UnicodeType, 0),
        3: ('matrix', p.UnicodeType, 0),
        4: ('mnemonic', p.UnicodeType, 0),
        5: ('node', HDNodeType, 0),
        6: ('passphrase_protection', p.BoolType, 0),
        7: ('reset_word', p.UnicodeType, 0),
        8: ('reset_entropy', p.BytesType, 0),
        9: ('recovery_fake_word', p.UnicodeType, 0),
        10: ('recovery_word_pos', p.UVarintType, 0),
        11: ('reset_word_pos', p.UVarintType, 0),
    }

    def __init__(
        self,
        layout: bytes = None,
        pin: str = None,
        matrix: str = None,
        mnemonic: str = None,
        node: HDNodeType = None,
        passphrase_protection: bool = None,
        reset_word: str = None,
        reset_entropy: bytes = None,
        recovery_fake_word: str = None,
        recovery_word_pos: int = None,
        reset_word_pos: int = None
    ) -> None:
        self.layout = layout
        self.pin = pin
        self.matrix = matrix
        self.mnemonic = mnemonic
        self.node = node
        self.passphrase_protection = passphrase_protection
        self.reset_word = reset_word
        self.reset_entropy = reset_entropy
        self.recovery_fake_word = recovery_fake_word
        self.recovery_word_pos = recovery_word_pos
        self.reset_word_pos = reset_word_pos



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from .common import SafeWISETest
from .conftest import SafeWISE_VERSION
from binascii import hexlify
from safewiselib import stellar
from safewiselib import messages as proto
from safewiselib.client import CallException
from safewiselib.tools import parse_path


@pytest.mark.stellar
@pytest.mark.xfail(SafeWISE_VERSION == 2, reason="T2 support is not yet finished")
class TestMsgStellarGetPublicKey(SafeWISETest):

    def test_stellar_get_public_key(self):
        self.setup_mnemonic_nopin_nopassphrase()

        # GAK5MSF74TJW6GLM7NLTL76YZJKM2S4CGP3UH4REJHPHZ4YBZW2GSBPW
        response = self.client.stellar_get_public_key(parse_path(stellar.DEFAULT_BIP32_PATH))
        assert hexlify(response) == b'15d648bfe4d36f196cfb5735ffd8ca54cd4b8233f743f22449de7cf301cdb469'
        assert stellar.address_from_public_key(response) == b'GAK5MSF74TJW6GLM7NLTL76YZJKM2S4CGP3UH4REJHPHZ4YBZW2GSBPW'

    def test_stellar_get_public_key_fail(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with pytest.raises(CallException) as exc:
            self.client.stellar_get_public_key(parse_path('m/0/1'))

        if SafeWISE_VERSION == 1:
            assert exc.value.args[0] == proto.FailureType.ProcessError
            assert exc.value.args[1].endswith('Failed to derive private key')
        else:
            assert exc.value.args[0] == proto.FailureType.FirmwareError
            assert exc.value.args[1].endswith('Firmware error')



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from binascii import unhexlify
import pytest

from .common import SafeWISETest
from safewiselib import messages as proto
from safewiselib.tools import parse_path

PUBLIC_KEY = unhexlify('eb56d7bbb5e8ea9269405f7a8527fe126023d1db2c973cfac6f760b60ae27294')


@pytest.mark.lisk
@pytest.mark.skip_t1
class TestMsgLiskSignTx(SafeWISETest):

    def test_lisk_sign_tx_send(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.LiskSignedTx(
                    signature=unhexlify('b62717d581e5713bca60b758b661e6cfa091addc6caedd57534e06cda805943ee80797b9fb9a1e1b2bd584e292d2a7f832a4d1b3f15f00e1ee1b72de7e195a08')
                )
            ])

            self.client.lisk_sign_tx(parse_path("m/44'/134'/0'/0'"), {
                "amount": "10000000",
                "recipientId": "9971262264659915921L",
                "timestamp": 57525937,
                "type": 0,
                "fee": "10000000",
                "asset": {}
            })

    def test_lisk_sign_tx_send_with_data(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.LiskSignedTx(
                    signature=unhexlify('5dd0dbb87ee46f3e985b1ef2df85cb0bec481e8601d150388f73e198cdd57a698eab076c7cd5b281fbb6a83dd3dc64d91a6eccd1614dffd46f101194ffa3a004')
                )
            ])

            self.client.lisk_sign_tx(parse_path("m/44'/134'/0'/0'"), {
                "amount": "10000000",
                "recipientId": "9971262264659915921L",
                "timestamp": 57525937,
                "type": 0,
                "fee": "20000000",
                "asset": {
                    "data": "Test data"
                }
            })

    def test_lisk_sign_tx_second_signature(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.PublicKey),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.LiskSignedTx(
                    signature=unhexlify('f02bdc40a7599c21d29db4080ff1ff8934f76eedf5b0c4fa695c8a64af2f0b40a5c4f92db203863eebbbfad8f0611a23f451ed8bb711490234cdfb034728fd01')
                )
            ])

            self.client.lisk_sign_tx(parse_path("m/44'/134'/0'/0'"), {
                "amount": "0",
                "timestamp": 57525937,
                "type": 1,
                "fee": "500000000",
                "asset": {
                    "signature": {
                        "publicKey": "5d036a858ce89f844491762eb89e2bfbd50a4a0a0da658e4b2628b25b117ae09"
                    }
                }
            })

    def test_lisk_sign_tx_delegate_registration(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.LiskSignedTx(
                    signature=unhexlify('5ac02b2882b9d7d0f944e48baadc27de1296cc08c3533f7c8e380fbbb9fb4a6ac81b5dc57060d7d8c68912eea24eb6e39024801bccc0d55020e2052b0c2bb701')
                )
            ])

            self.client.lisk_sign_tx(parse_path("m/44'/134'/0'/0'"), {
                "amount": "0",
                "timestamp": 57525937,
                "type": 2,
                "fee": "2500000000",
                "asset": {
                    "delegate": {
                        "username": "safewise_t"
                    }
                }
            })

    def test_lisk_sign_tx_cast_votes(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.LiskSignedTx(
                    signature=unhexlify('1d0599a8387edaa4a6d309b8a78accd1ceaff20ff9d87136b01cba0efbcb9781c13dc2b0bab5a1ea4f196d8dcc9dbdbd2d56dbffcc088fc77686b2e2c2fe560f')
                )
            ])

            self.client.lisk_sign_tx(parse_path("m/44'/134'/0'/0'"), {
                "amount": "0",
                "timestamp": 57525937,
                "type": 3,
                "fee": "100000000",
                "asset": {
                    "votes": [
                        "+b002f58531c074c7190714523eec08c48db8c7cfc0c943097db1a2e82ed87f84",
                        "-ec111c8ad482445cfe83d811a7edd1f1d2765079c99d7d958cca1354740b7614"
                    ]
                }
            })

    def test_lisk_sign_tx_multisignature(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            self.client.set_expected_responses([
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.LiskSignedTx(
                    signature=unhexlify('88923866c2d500a6927715699ab41a0f58ea4b52e552d90e923bc24ac9da240f2328c93f9ce043a1da4937d4b61c7f57c02fc931f9824d06b24731e7be23c506')
                )
            ])

            self.client.lisk_sign_tx(parse_path("m/44'/134'/0'/0'"), {
                "amount": "0",
                "timestamp": 57525937,
                "type": 4,
                "fee": "1500000000",
                "asset": {
                    "multisignature": {
                        "min": 2,
                        "lifetime": 5,
                        "keysgroup": [
                            "+5d036a858ce89f844491762eb89e2bfbd50a4a0a0da658e4b2628b25b117ae09",
                            "+922fbfdd596fa78269bbcadc67ec2a1cc15fc929a19c462169568d7a3df1a1aa"
                        ]
                    }
                }
            })



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import functools
import os
import pytest

from safewiselib.transport import get_transport
from safewiselib.client import SafeWISEClient, SafeWISEClientDebugLink
from safewiselib import log, coins


def get_device():
    path = os.environ.get('SafeWISE_PATH')
    return get_transport(path)


def device_version():
    device = get_device()
    if not device:
        raise RuntimeError()
    client = SafeWISEClient(device)
    if client.features.model == "T":
        return 2
    else:
        return 1


SafeWISE_VERSION = device_version()


@pytest.fixture(scope="function")
def client():
    wirelink = get_device()
    debuglink = wirelink.find_debug()
    client = SafeWISEClientDebugLink(wirelink)
    client.set_debuglink(debuglink)
    client.set_tx_api(coins.tx_api['Bitcoin'])
    client.wipe_device()
    client.transport.session_begin()

    yield client

    client.transport.session_end()


def setup_client(mnemonic=None, pin='', passphrase=False):
    if mnemonic is None:
        mnemonic = ' '.join(['all'] * 12)
    if pin is True:
        pin = '1234'

    def client_decorator(function):
        @functools.wraps(function)
        def wrapper(client, *args, **kwargs):
            client.load_device_by_mnemonic(mnemonic=mnemonic, pin=pin, passphrase_protection=passphrase, label='test', language='english')
            return function(client, *args, **kwargs)
        return wrapper

    return client_decorator


def pytest_configure(config):
    if config.getoption('verbose'):
        log.enable_debug_output()


def pytest_addoption(parser):
    parser.addini("run_xfail", "List of markers that will run even if marked as xfail", "args", [])


def pytest_runtest_setup(item):
    """
    Called for each test item (class, individual tests).

    Performs custom processing, mainly useful for safewise CI testing:
    * 'skip_t2' tests are skipped on T2 and 'skip_t1' tests are skipped on T1.
    * no test should have both skips at the same time
    * allows to 'runxfail' tests specified by 'run_xfail' in pytest.ini
    """
    if item.get_marker("skip_t1") and item.get_marker("skip_t2"):
        pytest.fail("Don't skip tests for both safewises!")

    if item.get_marker("skip_t2") and SafeWISE_VERSION == 2:
        pytest.skip("Test excluded on SafeWISE T")
    if item.get_marker("skip_t1") and SafeWISE_VERSION == 1:
        pytest.skip("Test excluded on SafeWISE 1")

    xfail = item.get_marker("xfail")
    run_xfail = any(item.get_marker(marker) for marker in item.config.getini("run_xfail"))
    if xfail and run_xfail:
        # Deep hack: pytest's private _evalxfail helper determines whether the test should xfail or not.
        # The helper caches its result even before this hook runs.
        # Here we force-set the result to False, meaning "test does NOT xfail, run as normal"
        # IOW, this is basically per-item "--runxfail"
        item._evalxfail.result = False



# Automatically generated by pb2py
from .. import protobuf as p


class DebugLinkLog(p.MessageType):
    MESSAGE_WIRE_TYPE = 104
    FIELDS = {
        1: ('level', p.UVarintType, 0),
        2: ('bucket', p.UnicodeType, 0),
        3: ('text', p.UnicodeType, 0),
    }

    def __init__(
        self,
        level: int = None,
        bucket: str = None,
        text: str = None
    ) -> None:
        self.level = level
        self.bucket = bucket
        self.text = text



#!/usr/bin/env python3
import sys
import os
import json

def get_tokens():
    tokens = []
    for s, i in networks:
        try:
            files = os.scandir('defs/ethereum/tokens/tokens/%s' % s)
        except FileNotFoundError:
            continue

        for f in files:
            if not f.path.endswith('.json'):
                continue

            # print('Processing', f.path)
            data = json.load(open(f.path, 'r'))
            data['chain'] = s
            data['chain_id'] = i
            tokens.append(data)

    return tokens


def print_tokens(tokens, python=False):
    for t in sorted(tokens, key=lambda x: x['chain'] + x['symbol'].upper()):
        address, name, symbol, decimal, chain, chain_id = t['address'], t['name'], t['symbol'], int(t['decimals']), t['chain'], t['chain_id']
        address = '\\x'.join([address[i:i + 2] for i in range(0, len(address), 2)])[2:].lower()
        if python:
            print("    (%d, b'%s', '%s', %d),  # %s / %s" % (chain_id, address, symbol, decimal, chain, name))
        else:
            print('\t{%2d, "%s", " %s", %d}, // %s / %s' % (chain_id, address, symbol, decimal, chain, name))

    return len(tokens)


# disabled are networks with no tokens defined in ethereum-lists/tokens repo

networks = [
    ('eth', 1),
    ('exp', 2),
    # ('rop', 3),
    # ('rin', 4),
    ('ubq', 8),
    # ('rsk', 30),
    # ('kov', 42),
    ('etc', 61),
]


def generate_c(tokens):
    count = print_tokens(tokens)
    print('-' * 32)
    print('#define TOKENS_COUNT %d' % count)


def generate_python(tokens):
    print('tokens = [')
    print_tokens(tokens, python=True)
    print(']')


if __name__ == '__main__':
    tokens = get_tokens()

    if len(sys.argv) > 1 and sys.argv[1] == '--python':
        generate_python(tokens)
    else:
        generate_c(tokens)



# Automatically generated by pb2py
from .. import protobuf as p


class TxRequestDetailsType(p.MessageType):
    FIELDS = {
        1: ('request_index', p.UVarintType, 0),
        2: ('tx_hash', p.BytesType, 0),
        3: ('extra_data_len', p.UVarintType, 0),
        4: ('extra_data_offset', p.UVarintType, 0),
    }

    def __init__(
        self,
        request_index: int = None,
        tx_hash: bytes = None,
        extra_data_len: int = None,
        extra_data_offset: int = None
    ) -> None:
        self.request_index = request_index
        self.tx_hash = tx_hash
        self.extra_data_len = extra_data_len
        self.extra_data_offset = extra_data_offset



# Automatically generated by pb2py
from .. import protobuf as p


class IdentityType(p.MessageType):
    FIELDS = {
        1: ('proto', p.UnicodeType, 0),
        2: ('user', p.UnicodeType, 0),
        3: ('host', p.UnicodeType, 0),
        4: ('port', p.UnicodeType, 0),
        5: ('path', p.UnicodeType, 0),
        6: ('index', p.UVarintType, 0),  # default=0
    }

    def __init__(
        self,
        proto: str = None,
        user: str = None,
        host: str = None,
        port: str = None,
        path: str = None,
        index: int = None
    ) -> None:
        self.proto = proto
        self.user = user
        self.host = host
        self.port = port
        self.path = path
        self.index = index



# Automatically generated by pb2py
from .. import protobuf as p


class EncryptedMessage(p.MessageType):
    MESSAGE_WIRE_TYPE = 50
    FIELDS = {
        1: ('nonce', p.BytesType, 0),
        2: ('message', p.BytesType, 0),
        3: ('hmac', p.BytesType, 0),
    }

    def __init__(
        self,
        nonce: bytes = None,
        message: bytes = None,
        hmac: bytes = None
    ) -> None:
        self.nonce = nonce
        self.message = message
        self.hmac = hmac



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from binascii import hexlify
import pytest

from .common import SafeWISETest


@pytest.mark.lisk
@pytest.mark.skip_t1
class TestMsgLiskGetPublicKey(SafeWISETest):

    def test_lisk_get_public_key(self):
        self.setup_mnemonic_nopin_nopassphrase()
        sig = self.client.lisk_get_public_key([2147483692, 2147483782, 2147483648, 2147483648])
        assert hexlify(sig.public_key) == b'eb56d7bbb5e8ea9269405f7a8527fe126023d1db2c973cfac6f760b60ae27294'



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import binascii
import json
from . import messages as proto

TYPE_TRANSACTION_TRANSFER = 0x0101
TYPE_IMPORTANCE_TRANSFER = 0x0801
TYPE_AGGREGATE_MODIFICATION = 0x1001
TYPE_MULTISIG_SIGNATURE = 0x1002
TYPE_MULTISIG = 0x1004
TYPE_PROVISION_NAMESPACE = 0x2001
TYPE_MOSAIC_CREATION = 0x4001
TYPE_MOSAIC_SUPPLY_CHANGE = 0x4002


def create_transaction_common(transaction):
    msg = proto.NEMTransactionCommon()
    msg.network = (transaction["version"] >> 24) & 0xFF
    msg.timestamp = transaction["timeStamp"]
    msg.fee = transaction["fee"]
    msg.deadline = transaction["deadline"]

    if "signer" in transaction:
        msg.signer = binascii.unhexlify(transaction["signer"])

    return msg


def create_transfer(transaction):
    msg = proto.NEMTransfer()
    msg.recipient = transaction["recipient"]
    msg.amount = transaction["amount"]

    if "payload" in transaction["message"]:
        msg.payload = binascii.unhexlify(transaction["message"]["payload"])

        if transaction["message"]["type"] == 0x02:
            msg.public_key = binascii.unhexlify(transaction["message"]["publicKey"])

    if "mosaics" in transaction:
        msg.mosaics = [proto.NEMMosaic(
            namespace=mosaic["mosaicId"]["namespaceId"],
            mosaic=mosaic["mosaicId"]["name"],
            quantity=mosaic["quantity"],
        ) for mosaic in transaction["mosaics"]]

    return msg


def create_aggregate_modification(transactions):
    msg = proto.NEMAggregateModification()
    msg.modifications = [proto.NEMCosignatoryModification(
        type=modification["modificationType"],
        public_key=binascii.unhexlify(modification["cosignatoryAccount"]),
    ) for modification in transactions["modifications"]]

    if "minCosignatories" in transactions:
        msg.relative_change = transactions["minCosignatories"]["relativeChange"]

    return msg


def create_provision_namespace(transaction):
    msg = proto.NEMProvisionNamespace()
    msg.namespace = transaction["newPart"]

    if transaction["parent"]:
        msg.parent = transaction["parent"]

    msg.sink = transaction["rentalFeeSink"]
    msg.fee = transaction["rentalFee"]
    return msg


def create_mosaic_creation(transaction):
    definition = transaction["mosaicDefinition"]
    msg = proto.NEMMosaicCreation()
    msg.definition = proto.NEMMosaicDefinition()
    msg.definition.namespace = definition["id"]["namespaceId"]
    msg.definition.mosaic = definition["id"]["name"]

    if definition["levy"]:
        msg.definition.levy = definition["levy"]["type"]
        msg.definition.fee = definition["levy"]["fee"]
        msg.definition.levy_address = definition["levy"]["recipient"]
        msg.definition.levy_namespace = definition["levy"]["mosaicId"]["namespaceId"]
        msg.definition.levy_mosaic = definition["levy"]["mosaicId"]["name"]

    msg.definition.description = definition["description"]

    for property in definition["properties"]:
        name = property["name"]
        value = json.loads(property["value"])

        if name == "divisibility":
            msg.definition.divisibility = value
        elif name == "initialSupply":
            msg.definition.supply = value
        elif name == "supplyMutable":
            msg.definition.mutable_supply = value
        elif name == "transferable":
            msg.definition.transferable = value

    msg.sink = transaction["creationFeeSink"]
    msg.fee = transaction["creationFee"]
    return msg


def create_supply_change(transaction):
    msg = proto.NEMMosaicSupplyChange()
    msg.namespace = transaction["mosaicId"]["namespaceId"]
    msg.mosaic = transaction["mosaicId"]["name"]
    msg.type = transaction["supplyType"]
    msg.delta = transaction["delta"]
    return msg


def create_importance_transfer(transaction):
    msg = proto.NEMImportanceTransfer()
    msg.mode = transaction["importanceTransfer"]["mode"]
    msg.public_key = binascii.unhexlify(transaction["importanceTransfer"]["publicKey"])
    return msg


def create_sign_tx(transaction):
    msg = proto.NEMSignTx()
    msg.transaction = create_transaction_common(transaction)
    msg.cosigning = (transaction["type"] == TYPE_MULTISIG_SIGNATURE)

    if transaction["type"] in (TYPE_MULTISIG_SIGNATURE, TYPE_MULTISIG):
        transaction = transaction["otherTrans"]
        msg.multisig = create_transaction_common(transaction)
    elif "otherTrans" in transaction:
        raise ValueError("Transaction does not support inner transaction")

    if transaction["type"] == TYPE_TRANSACTION_TRANSFER:
        msg.transfer = create_transfer(transaction)
    elif transaction["type"] == TYPE_AGGREGATE_MODIFICATION:
        msg.aggregate_modification = create_aggregate_modification(transaction)
    elif transaction["type"] == TYPE_PROVISION_NAMESPACE:
        msg.provision_namespace = create_provision_namespace(transaction)
    elif transaction["type"] == TYPE_MOSAIC_CREATION:
        msg.mosaic_creation = create_mosaic_creation(transaction)
    elif transaction["type"] == TYPE_MOSAIC_SUPPLY_CHANGE:
        msg.supply_change = create_supply_change(transaction)
    elif transaction["type"] == TYPE_IMPORTANCE_TRANSFER:
        msg.importance_transfer = create_importance_transfer(transaction)
    else:
        raise ValueError("Unknown transaction type")

    return msg



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import os
import socket

from ..protocol_v1 import ProtocolV1
from ..protocol_v2 import ProtocolV2
from . import Transport, TransportException


class UdpTransport(Transport):

    DEFAULT_HOST = '127.0.0.1'
    DEFAULT_PORT = 21324
    PATH_PREFIX = 'udp'

    def __init__(self, device=None, protocol=None):
        super(UdpTransport, self).__init__()

        if not device:
            host = UdpTransport.DEFAULT_HOST
            port = UdpTransport.DEFAULT_PORT
        else:
            devparts = device.split(':')
            host = devparts[0]
            port = int(devparts[1]) if len(devparts) > 1 else UdpTransport.DEFAULT_PORT
        if not protocol:
            protocol = ProtocolV1()
        self.device = (host, port)
        self.protocol = protocol
        self.socket = None

    def get_path(self):
        return "%s:%s:%s" % ((self.PATH_PREFIX,) + self.device)

    def find_debug(self):
        host, port = self.device
        return UdpTransport('{}:{}'.format(host, port + 1), self.protocol)

    @classmethod
    def _try_path(cls, path):
        d = cls(path)
        try:
            d.open()
            if d._ping():
                return d
            else:
                raise TransportException('No SafeWISE device found at address {}'.format(path))
        finally:
            d.close()

    @classmethod
    def enumerate(cls):
        default_path = '{}:{}'.format(cls.DEFAULT_HOST, cls.DEFAULT_PORT)
        try:
            return [cls._try_path(default_path)]
        except TransportException:
            return []

    @classmethod
    def find_by_path(cls, path, prefix_search=False):
        if prefix_search:
            return super().find_by_path(path, prefix_search)
        else:
            path = path.replace('{}:'.format(cls.PATH_PREFIX), '')
            return cls._try_path(path)

    def open(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.connect(self.device)
        self.socket.settimeout(10)
        self.protocol.session_begin(self)

    def close(self):
        if self.socket:
            self.protocol.session_end(self)
            self.socket.close()
            self.socket = None

    def _ping(self):
        '''Test if the device is listening.'''
        resp = None
        try:
            self.socket.sendall(b'PINGPING')
            resp = self.socket.recv(8)
        except:
            pass
        return resp == b'PONGPONG'

    def read(self):
        return self.protocol.read(self)

    def write(self, msg):
        return self.protocol.write(self, msg)

    def write_chunk(self, chunk):
        if len(chunk) != 64:
            raise TransportException('Unexpected data length')
        self.socket.sendall(chunk)

    def read_chunk(self):
        while True:
            try:
                chunk = self.socket.recv(64)
                break
            except socket.timeout:
                continue
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        return bytearray(chunk)


TRANSPORT = UdpTransport



# Automatically generated by pb2py
from .. import protobuf as p


class ButtonRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 26
    FIELDS = {
        1: ('code', p.UVarintType, 0),
        2: ('data', p.UnicodeType, 0),
    }

    def __init__(
        self,
        code: int = None,
        data: str = None
    ) -> None:
        self.code = code
        self.data = data



# Automatically generated by pb2py
from .. import protobuf as p


class EthereumTxAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 60
    FIELDS = {
        1: ('data_chunk', p.BytesType, 0),
    }

    def __init__(
        self,
        data_chunk: bytes = None
    ) -> None:
        self.data_chunk = data_chunk



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import time
import hid
import os
import sys

from ..protocol_v1 import ProtocolV1
from ..protocol_v2 import ProtocolV2
from . import Transport, TransportException

DEV_SafeWISE1 = (0x1209, 0xabba)
DEV_SafeWISE2 = (0xffff, 0x53c1)
DEV_SafeWISE2_BL = (0xffff, 0x53c0)


class HidHandle:

    def __init__(self, path):
        self.path = path
        self.count = 0
        self.handle = None

    def open(self):
        if self.count == 0:
            self.handle = hid.device()
            try:
                self.handle.open_path(self.path)
            except (IOError, OSError) as e:
                if sys.platform.startswith('linux'):
                    e.args = e.args + ('Do you have udev rules installed? https://github.com/trezor/trezor-common/blob/master/udev/51-trezor.rules', )
                raise e
            self.handle.set_nonblocking(True)
        self.count += 1

    def close(self):
        if self.count == 1:
            self.handle.close()
        if self.count > 0:
            self.count -= 1


class HidTransport(Transport):
    '''
    HidTransport implements transport over USB HID interface.
    '''

    PATH_PREFIX = 'hid'

    def __init__(self, device, protocol=None, hid_handle=None):
        super(HidTransport, self).__init__()

        if hid_handle is None:
            hid_handle = HidHandle(device['path'])

        if protocol is None:
            # force_v1 = os.environ.get('SafeWISE_TRANSPORT_V1', '0')
            force_v1 = True

            if is_safewise2(device) and not int(force_v1):
                protocol = ProtocolV2()
            else:
                protocol = ProtocolV1()

        self.device = device
        self.protocol = protocol
        self.hid = hid_handle
        self.hid_version = None

    def get_path(self):
        return "%s:%s" % (self.PATH_PREFIX, self.device['path'].decode())

    @staticmethod
    def enumerate(debug=False):
        devices = []
        for dev in hid.enumerate(0, 0):
            if not (is_safewise1(dev) or is_safewise2(dev) or is_safewise2_bl(dev)):
                continue
            if debug:
                if not is_debuglink(dev):
                    continue
            else:
                if not is_wirelink(dev):
                    continue
            devices.append(HidTransport(dev))
        return devices

    def find_debug(self):
        if isinstance(self.protocol, ProtocolV2):
            # For v2 protocol, lets use the same HID interface, but with a different session
            protocol = ProtocolV2()
            debug = HidTransport(self.device, protocol, self.hid)
            return debug
        if isinstance(self.protocol, ProtocolV1):
            # For v1 protocol, find debug USB interface for the same serial number
            for debug in HidTransport.enumerate(debug=True):
                if debug.device['serial_number'] == self.device['serial_number']:
                    return debug
        raise TransportException('Debug HID device not found')

    def open(self):
        self.hid.open()
        if is_safewise1(self.device):
            self.hid_version = self.probe_hid_version()
        else:
            self.hid_version = 2
        self.protocol.session_begin(self)

    def close(self):
        self.protocol.session_end(self)
        self.hid.close()
        self.hid_version = None

    def read(self):
        return self.protocol.read(self)

    def write(self, msg):
        return self.protocol.write(self, msg)

    def write_chunk(self, chunk):
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        if self.hid_version == 2:
            self.hid.handle.write(b'\0' + bytearray(chunk))
        else:
            self.hid.handle.write(chunk)

    def read_chunk(self):
        while True:
            chunk = self.hid.handle.read(64)
            if chunk:
                break
            else:
                time.sleep(0.001)
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        return bytearray(chunk)

    def probe_hid_version(self):
        n = self.hid.handle.write([0, 63] + [0xFF] * 63)
        if n == 65:
            return 2
        n = self.hid.handle.write([63] + [0xFF] * 63)
        if n == 64:
            return 1
        raise TransportException('Unknown HID version')


def is_safewise1(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_SafeWISE1


def is_safewise2(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_SafeWISE2


def is_safewise2_bl(dev):
    return (dev['vendor_id'], dev['product_id']) == DEV_SafeWISE2_BL


def is_wirelink(dev):
    return (dev['usage_page'] == 0xFF00 or dev['interface_number'] == 0)


def is_debuglink(dev):
    return (dev['usage_page'] == 0xFF01 or dev['interface_number'] == 1)


TRANSPORT = HidTransport



# Automatically generated by pb2py
from .. import protobuf as p


class RecoveryDevice(p.MessageType):
    MESSAGE_WIRE_TYPE = 45
    FIELDS = {
        1: ('word_count', p.UVarintType, 0),
        2: ('passphrase_protection', p.BoolType, 0),
        3: ('pin_protection', p.BoolType, 0),
        4: ('language', p.UnicodeType, 0),  # default='english'
        5: ('label', p.UnicodeType, 0),
        6: ('enforce_wordlist', p.BoolType, 0),
        8: ('type', p.UVarintType, 0),
        9: ('u2f_counter', p.UVarintType, 0),
        10: ('dry_run', p.BoolType, 0),
    }

    def __init__(
        self,
        word_count: int = None,
        passphrase_protection: bool = None,
        pin_protection: bool = None,
        language: str = None,
        label: str = None,
        enforce_wordlist: bool = None,
        type: int = None,
        u2f_counter: int = None,
        dry_run: bool = None
    ) -> None:
        self.word_count = word_count
        self.passphrase_protection = passphrase_protection
        self.pin_protection = pin_protection
        self.language = language
        self.label = label
        self.enforce_wordlist = enforce_wordlist
        self.type = type
        self.u2f_counter = u2f_counter
        self.dry_run = dry_run



# Automatically generated by pb2py
from .. import protobuf as p


class TxRequestSerializedType(p.MessageType):
    FIELDS = {
        1: ('signature_index', p.UVarintType, 0),
        2: ('signature', p.BytesType, 0),
        3: ('serialized_tx', p.BytesType, 0),
    }

    def __init__(
        self,
        signature_index: int = None,
        signature: bytes = None,
        serialized_tx: bytes = None
    ) -> None:
        self.signature_index = signature_index
        self.signature = signature
        self.serialized_tx = serialized_tx



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from .common import SafeWISETest
from safewiselib.tools import parse_path


@pytest.mark.nem
class TestMsgNEMGetaddress(SafeWISETest):

    def test_nem_getaddress(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert self.client.nem_get_address(parse_path("m/44'/1'/0'/0'/0'"), 0x68) == "NB3JCHVARQNGDS3UVGAJPTFE22UQFGMCQGHUBWQN"
        assert self.client.nem_get_address(parse_path("m/44'/1'/0'/0'/0'"), 0x98) == "TB3JCHVARQNGDS3UVGAJPTFE22UQFGMCQHSBNBMF"



# Automatically generated by pb2py
from .. import protobuf as p


class LiskDelegateType(p.MessageType):
    FIELDS = {
        1: ('username', p.UnicodeType, 0),
    }

    def __init__(
        self,
        username: str = None
    ) -> None:
        self.username = username



# Automatically generated by pb2py
from .. import protobuf as p


class EthereumVerifyMessage(p.MessageType):
    MESSAGE_WIRE_TYPE = 65
    FIELDS = {
        1: ('address', p.BytesType, 0),
        2: ('signature', p.BytesType, 0),
        3: ('message', p.BytesType, 0),
    }

    def __init__(
        self,
        address: bytes = None,
        signature: bytes = None,
        message: bytes = None
    ) -> None:
        self.address = address
        self.signature = signature
        self.message = message



# Automatically generated by pb2py
from .. import protobuf as p


class ApplySettings(p.MessageType):
    MESSAGE_WIRE_TYPE = 25
    FIELDS = {
        1: ('language', p.UnicodeType, 0),
        2: ('label', p.UnicodeType, 0),
        3: ('use_passphrase', p.BoolType, 0),
        4: ('homescreen', p.BytesType, 0),
        5: ('passphrase_source', p.UVarintType, 0),
        6: ('auto_lock_delay_ms', p.UVarintType, 0),
    }

    def __init__(
        self,
        language: str = None,
        label: str = None,
        use_passphrase: bool = None,
        homescreen: bytes = None,
        passphrase_source: int = None,
        auto_lock_delay_ms: int = None
    ) -> None:
        self.language = language
        self.label = label
        self.use_passphrase = use_passphrase
        self.homescreen = homescreen
        self.passphrase_source = passphrase_source
        self.auto_lock_delay_ms = auto_lock_delay_ms



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import importlib
import logging

from typing import Iterable, Type, List, Set

LOG = logging.getLogger(__name__)


class TransportException(Exception):
    pass


class Transport(object):

    def __init__(self):
        self.session_counter = 0

    def __str__(self):
        return self.get_path()

    def get_path(self):
        return '{}:{}'.format(self.PATH_PREFIX, self.device)

    def session_begin(self):
        if self.session_counter == 0:
            self.open()
        self.session_counter += 1

    def session_end(self):
        self.session_counter = max(self.session_counter - 1, 0)
        if self.session_counter == 0:
            self.close()

    def open(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    @classmethod
    def enumerate(cls):
        raise NotImplementedError

    @classmethod
    def find_by_path(cls, path, prefix_search=False):
        for device in cls.enumerate():
            if path is None or device.get_path() == path \
                    or (prefix_search and device.get_path().startswith(path)):
                return device

        raise TransportException('{} device not found: {}'.format(cls.PATH_PREFIX, path))


def all_transports() -> Iterable[Type[Transport]]:
    transports = set()  # type: Set[Type[Transport]]
    for modname in ("bridge", "hid", "udp", "webusb"):
        try:
            # Import the module and find the Transport class.
            # To avoid iterating over every item, the module should assign its Transport class
            # to a constant named TRANSPORT.
            module = importlib.import_module("." + modname, __name__)
            try:
                transports.add(getattr(module, "TRANSPORT"))
            except AttributeError:
                LOG.warning("Skipping broken module {}".format(modname))
        except ImportError as e:
            LOG.info("Failed to import module {}: {}".format(modname, e))

    return transports


def enumerate_devices() -> Iterable[Transport]:
    devices = []  # type: List[Transport]
    for transport in all_transports():
        try:
            found = transport.enumerate()
            LOG.info("Enumerating {}: found {} devices".format(transport.__name__, len(found)))
            devices.extend(found)
        except NotImplementedError:
            LOG.error("{} does not implement device enumeration".format(transport.__name__))
        except Exception as e:
            LOG.error("Failed to enumerate {}. {}: {}".format(transport.__name__, e.__class__.__name__, e))
    return devices


def get_transport(path: str = None, prefix_search: bool = False) -> Transport:
    if path is None:
        try:
            return next(iter(enumerate_devices()))
        except IndexError:
            raise Exception("No SafeWISE device found") from None

    # Find whether B is prefix of A (transport name is part of the path)
    # or A is prefix of B (path is a prefix, or a name, of transport).
    # This naively expects that no two transports have a common prefix.
    def match_prefix(a: str, b: str) -> bool:
        return a.startswith(b) or b.startswith(a)

    LOG.info("looking for device by {}: {}".format("prefix" if prefix_search else "full path", path))
    transports = [t for t in all_transports() if match_prefix(path, t.PATH_PREFIX)]
    if transports:
        return transports[0].find_by_path(path, prefix_search=prefix_search)

    raise Exception("Could not find device by path: {}".format(path))



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from .common import SafeWISETest


@pytest.mark.lisk
@pytest.mark.skip_t1
class TestMsgLiskGetaddress(SafeWISETest):

    def test_lisk_getaddress(self):
        self.setup_mnemonic_nopin_nopassphrase()
        assert self.client.lisk_get_address([2147483692, 2147483782]) == '1431530009238518937L'
        assert self.client.lisk_get_address([2147483692, 2147483782, 2147483648]) == '17563781916205589679L'
        assert self.client.lisk_get_address([2147483692, 2147483782, 2147483648, 2147483649]) == '1874186517773691964L'
        assert self.client.lisk_get_address([2147483692, 2147483782, 2147484647, 2147484647]) == '16295203558710684671L'



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import logging
from typing import Set, Type, Optional

from . import protobuf

OMITTED_MESSAGES = set()  # type: Set[Type[protobuf.MessageType]]


class PrettyProtobufFormatter(logging.Formatter):

    def format(self, record: logging.LogRecord) -> str:
        time = self.formatTime(record)
        message = '[{time}] {source} {level}: {msg}'.format(
            time=time, level=record.levelname.upper(),
            source=record.name,
            msg=super().format(record))
        if hasattr(record, 'protobuf'):
            if type(record.protobuf) in OMITTED_MESSAGES:
                message += " ({} bytes)".format(record.protobuf.ByteSize())
            else:
                message += "\n" + protobuf.format_message(record.protobuf)
        return message


def enable_debug_output(handler: Optional[logging.Handler] = None):
    if handler is None:
        handler = logging.StreamHandler()

    formatter = PrettyProtobufFormatter()
    handler.setFormatter(formatter)

    logger = logging.getLogger('safewiselib')
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)



# Automatically generated by pb2py
from .. import protobuf as p


class FirmwareErase(p.MessageType):
    MESSAGE_WIRE_TYPE = 6
    FIELDS = {
        1: ('length', p.UVarintType, 0),
    }

    def __init__(
        self,
        length: int = None
    ) -> None:
        self.length = length



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from .common import SafeWISETest
from ..support import ckd_public as bip32
from safewiselib import messages as proto
from safewiselib.tools import parse_path


class TestMsgGetaddressSegwitNative(SafeWISETest):

    def test_show_segwit(self):
        self.setup_mnemonic_allallall()
        assert self.client.get_address("Testnet", parse_path("49'/1'/0'/0/0"), True, None, script_type=proto.InputScriptType.SPENDWITNESS) == 'tb1qqzv60m9ajw8drqulta4ld4gfx0rdh82un5s65s'
        assert self.client.get_address("Testnet", parse_path("49'/1'/0'/1/0"), False, None, script_type=proto.InputScriptType.SPENDWITNESS) == 'tb1q694ccp5qcc0udmfwgp692u2s2hjpq5h407urtu'
        assert self.client.get_address("Testnet", parse_path("44'/1'/0'/0/0"), False, None, script_type=proto.InputScriptType.SPENDWITNESS) == 'tb1q54un3q39sf7e7tlfq99d6ezys7qgc62a6rxllc'
        assert self.client.get_address("Testnet", parse_path("44'/1'/0'/0/0"), False, None, script_type=proto.InputScriptType.SPENDADDRESS) == 'mvbu1Gdy8SUjTenqerxUaZyYjmveZvt33q'

    def test_show_multisig_3(self):
        self.setup_mnemonic_allallall()
        nodes = [self.client.get_public_node(parse_path("999'/1'/%d'" % index)) for index in range(1, 4)]
        multisig1 = proto.MultisigRedeemScriptType(
            pubkeys=list(map(lambda n: proto.HDNodePathType(node=bip32.deserialize(n.xpub), address_n=[2, 0]), nodes)),
            signatures=[b'', b'', b''],
            m=2,
        )
        multisig2 = proto.MultisigRedeemScriptType(
            pubkeys=list(map(lambda n: proto.HDNodePathType(node=bip32.deserialize(n.xpub), address_n=[2, 1]), nodes)),
            signatures=[b'', b'', b''],
            m=2,
        )
        for i in [1, 2, 3]:
            assert self.client.get_address("Testnet", parse_path("999'/1'/%d'/2/1" % i), False, multisig2, script_type=proto.InputScriptType.SPENDWITNESS) == 'tb1qch62pf820spe9mlq49ns5uexfnl6jzcezp7d328fw58lj0rhlhasge9hzy'
            assert self.client.get_address("Testnet", parse_path("999'/1'/%d'/2/0" % i), False, multisig1, script_type=proto.InputScriptType.SPENDWITNESS) == 'tb1qr6xa5v60zyt3ry9nmfew2fk5g9y3gerkjeu6xxdz7qga5kknz2ssld9z2z'



# Automatically generated by pb2py
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None
from .NEMMosaic import NEMMosaic


class NEMTransfer(p.MessageType):
    FIELDS = {
        1: ('recipient', p.UnicodeType, 0),
        2: ('amount', p.UVarintType, 0),
        3: ('payload', p.BytesType, 0),
        4: ('public_key', p.BytesType, 0),
        5: ('mosaics', NEMMosaic, p.FLAG_REPEATED),
    }

    def __init__(
        self,
        recipient: str = None,
        amount: int = None,
        payload: bytes = None,
        public_key: bytes = None,
        mosaics: List[NEMMosaic] = None
    ) -> None:
        self.recipient = recipient
        self.amount = amount
        self.payload = payload
        self.public_key = public_key
        self.mosaics = mosaics if mosaics is not None else []



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import mock

from safewiselib.transport import all_transports


def test_all_transports_without_hid():
    # import all transports, assume this doesn't fail
    transports_ref = all_transports()
    # also shouldn't fail when bridge transport is missing
    with mock.patch.dict('sys.modules', {'safewiselib.transport.bridge': None}):
        transports = all_transports()
        # there should now be less transports
        assert len(transports_ref) > len(transports)



# Automatically generated by pb2py
from .. import protobuf as p


class SignedIdentity(p.MessageType):
    MESSAGE_WIRE_TYPE = 54
    FIELDS = {
        1: ('address', p.UnicodeType, 0),
        2: ('public_key', p.BytesType, 0),
        3: ('signature', p.BytesType, 0),
    }

    def __init__(
        self,
        address: str = None,
        public_key: bytes = None,
        signature: bytes = None
    ) -> None:
        self.address = address
        self.public_key = public_key
        self.signature = signature



# Automatically generated by pb2py
from .. import protobuf as p
from .StellarAssetType import StellarAssetType


class StellarCreatePassiveOfferOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 214
    FIELDS = {
        1: ('source_account', p.BytesType, 0),
        2: ('selling_asset', StellarAssetType, 0),
        3: ('buying_asset', StellarAssetType, 0),
        4: ('amount', p.SVarintType, 0),
        5: ('price_n', p.UVarintType, 0),
        6: ('price_d', p.UVarintType, 0),
    }

    def __init__(
        self,
        source_account: bytes = None,
        selling_asset: StellarAssetType = None,
        buying_asset: StellarAssetType = None,
        amount: int = None,
        price_n: int = None,
        price_d: int = None
    ) -> None:
        self.source_account = source_account
        self.selling_asset = selling_asset
        self.buying_asset = buying_asset
        self.amount = amount
        self.price_n = price_n
        self.price_d = price_d



#!/usr/bin/env python3
from safewiselib.debuglink import DebugLink
from safewiselib.client import SafeWISEClient
from safewiselib.transport import enumerate_devices
import binascii
import sys

sectoraddrs = [0x8000000, 0x8004000, 0x8008000, 0x800c000,
               0x8010000, 0x8020000, 0x8040000, 0x8060000,
               0x8080000, 0x80a0000, 0x80c0000, 0x80f0000]
sectorlens = [0x4000, 0x4000, 0x4000, 0x4000,
              0x8000, 0x10000, 0x10000, 0x10000,
              0x10000, 0x10000, 0x10000, 0x10000]


def main():
    # List all debuggable SafeWISEs
    devices = [device for device in enumerate_devices() if hasattr(device, 'find_debug')]

    # Check whether we found any
    if len(devices) == 0:
        print('No SafeWISE found')
        return

    # Use first connected device
    transport = devices[0]
    debug_transport = devices[0].find_debug()

    # Creates object for manipulating SafeWISE
    client = SafeWISEClient(transport)
    debug = DebugLink(debug_transport)

    sector = int(sys.argv[1])
    f = open(sys.argv[2], "rb")
    content = f.read(sectorlens[sector])
    if (len(content) != sectorlens[sector]):
        print("Not enough bytes in file")
        return

    debug.flash_erase(sector)
    step = 0x400
    for offset in range(0, sectorlens[sector], step):
        debug.memory_write(sectoraddrs[sector] + offset, content[offset:offset + step], flash=True)
    client.close()


if __name__ == '__main__':
    main()



# Automatically generated by pb2py
from .. import protobuf as p


class Success(p.MessageType):
    MESSAGE_WIRE_TYPE = 2
    FIELDS = {
        1: ('message', p.UnicodeType, 0),
    }

    def __init__(
        self,
        message: str = None
    ) -> None:
        self.message = message



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import time
import pytest

from .common import SafeWISETest, generate_entropy

from safewiselib import messages as proto
from mnemonic import Mnemonic


@pytest.mark.skip_t1
class TestMsgResetDeviceT2(SafeWISETest):

    def test_reset_device(self):

        # No PIN, no passphrase, don't display random
        external_entropy = b'zlutoucky kun upel divoke ody' * 2
        strength = 128
        ret = self.client.call_raw(proto.ResetDevice(
            display_random=False,
            strength=strength,
            passphrase_protection=False,
            pin_protection=False,
            label='test'
        ))

        # Provide entropy
        assert isinstance(ret, proto.EntropyRequest)
        internal_entropy = self.client.debug.read_reset_entropy()
        ret = self.client.call_raw(proto.EntropyAck(entropy=external_entropy))

        # Generate mnemonic locally
        entropy = generate_entropy(strength, internal_entropy, external_entropy)
        expected_mnemonic = Mnemonic('english').to_mnemonic(entropy)

        # Safety warning
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        # List through mnemonic pages
        assert isinstance(ret, proto.ButtonRequest)
        self.client.transport.write(proto.ButtonAck())
        self.client.debug.press_yes()
        words = []
        time.sleep(1)
        words.extend(self.client.debug.read_reset_word().split())
        self.client.debug.swipe_down()
        time.sleep(1)
        words.extend(self.client.debug.read_reset_word().split())
        self.client.debug.swipe_down()
        time.sleep(1)
        words.extend(self.client.debug.read_reset_word().split())

        # Compare that device generated proper mnemonic for given entropies
        assert ' '.join(words) == expected_mnemonic

        # Confirm the mnemonic
        self.client.debug.press_yes()

        # Check mnemonic words
        time.sleep(1)
        index = self.client.debug.read_reset_word_pos()
        self.client.debug.input(words[index])
        time.sleep(1)
        index = self.client.debug.read_reset_word_pos()
        self.client.debug.input(words[index])
        ret = self.client.transport.read()

        # Safety warning
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())
        assert isinstance(ret, proto.Success)

        # Check if device is properly initialized
        resp = self.client.call_raw(proto.Initialize())
        assert resp.initialized is True
        assert resp.needs_backup is False
        assert resp.pin_protection is False
        assert resp.passphrase_protection is False

    def test_reset_device_pin(self):

        # PIN, passphrase, display random
        external_entropy = b'zlutoucky kun upel divoke ody' * 2
        strength = 128
        ret = self.client.call_raw(proto.ResetDevice(
            display_random=True,
            strength=strength,
            passphrase_protection=True,
            pin_protection=True,
            label='test'
        ))

        # Enter PIN for first time
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.input('654')
        ret = self.client.call_raw(proto.ButtonAck())

        # Enter PIN for second time
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.input('654')
        ret = self.client.call_raw(proto.ButtonAck())

        # Confirm entropy
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        # Provide entropy
        assert isinstance(ret, proto.EntropyRequest)
        internal_entropy = self.client.debug.read_reset_entropy()
        ret = self.client.call_raw(proto.EntropyAck(entropy=external_entropy))

        # Generate mnemonic locally
        entropy = generate_entropy(strength, internal_entropy, external_entropy)
        expected_mnemonic = Mnemonic('english').to_mnemonic(entropy)

        # Safety warning
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())

        # List through mnemonic pages
        assert isinstance(ret, proto.ButtonRequest)
        self.client.transport.write(proto.ButtonAck())
        self.client.debug.press_yes()
        words = []
        time.sleep(1)
        words.extend(self.client.debug.read_reset_word().split())
        self.client.debug.swipe_down()
        time.sleep(1)
        words.extend(self.client.debug.read_reset_word().split())
        self.client.debug.swipe_down()
        time.sleep(1)
        words.extend(self.client.debug.read_reset_word().split())

        # Compare that device generated proper mnemonic for given entropies
        assert ' '.join(words) == expected_mnemonic

        # Confirm the mnemonic
        self.client.debug.press_yes()

        # Check mnemonic words
        time.sleep(1)
        index = self.client.debug.read_reset_word_pos()
        self.client.debug.input(words[index])
        time.sleep(1)
        index = self.client.debug.read_reset_word_pos()
        self.client.debug.input(words[index])
        ret = self.client.transport.read()

        # Safety warning
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        ret = self.client.call_raw(proto.ButtonAck())
        assert isinstance(ret, proto.Success)

        # Check if device is properly initialized
        resp = self.client.call_raw(proto.Initialize())
        assert resp.initialized is True
        assert resp.needs_backup is False
        assert resp.pin_protection is True
        assert resp.passphrase_protection is True

    def test_failed_pin(self):
        # external_entropy = b'zlutoucky kun upel divoke ody' * 2
        strength = 128
        ret = self.client.call_raw(proto.ResetDevice(
            strength=strength,
            pin_protection=True,
            label='test'
        ))

        # Enter PIN for first time
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.input('654')
        ret = self.client.call_raw(proto.ButtonAck())

        # Enter PIN for second time
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.input('456')
        ret = self.client.call_raw(proto.ButtonAck())

        assert isinstance(ret, proto.ButtonRequest)

    def test_already_initialized(self):
        self.setup_mnemonic_nopin_nopassphrase()
        with pytest.raises(Exception):
            self.client.reset_device(False, 128, True, True, 'label', 'english')



# Automatically generated by pb2py

from .HDNodePathType import HDNodePathType
from .HDNodeType import HDNodeType
from .IdentityType import IdentityType
from .LiskDelegateType import LiskDelegateType
from .LiskMultisignatureType import LiskMultisignatureType
from .LiskSignatureType import LiskSignatureType
from .LiskTransactionAsset import LiskTransactionAsset
from .LiskTransactionCommon import LiskTransactionCommon
from .MultisigRedeemScriptType import MultisigRedeemScriptType
from .NEMAggregateModification import NEMAggregateModification
from .NEMCosignatoryModification import NEMCosignatoryModification
from .NEMImportanceTransfer import NEMImportanceTransfer
from .NEMMosaic import NEMMosaic
from .NEMMosaicCreation import NEMMosaicCreation
from .NEMMosaicDefinition import NEMMosaicDefinition
from .NEMMosaicSupplyChange import NEMMosaicSupplyChange
from .NEMProvisionNamespace import NEMProvisionNamespace
from .NEMTransactionCommon import NEMTransactionCommon
from .NEMTransfer import NEMTransfer
from .StellarAssetType import StellarAssetType
from .TransactionType import TransactionType
from .TxInputType import TxInputType
from .TxOutputBinType import TxOutputBinType
from .TxOutputType import TxOutputType
from .TxRequestDetailsType import TxRequestDetailsType
from .TxRequestSerializedType import TxRequestSerializedType
from . import FailureType
from . import OutputScriptType
from . import InputScriptType
from . import RequestType
from . import ButtonRequestType
from . import PinMatrixRequestType
from . import RecoveryDeviceType
from . import WordRequestType
from . import PassphraseSourceType
from . import NEMMosaicLevy
from . import NEMSupplyChangeType
from . import NEMModificationType
from . import NEMImportanceTransferMode
from . import LiskTransactionType
from .Address import Address
from .ApplyFlags import ApplyFlags
from .ApplySettings import ApplySettings
from .BackupDevice import BackupDevice
from .ButtonAck import ButtonAck
from .ButtonRequest import ButtonRequest
from .Cancel import Cancel
from .ChangePin import ChangePin
from .CipherKeyValue import CipherKeyValue
from .CipheredKeyValue import CipheredKeyValue
from .ClearSession import ClearSession
from .CosiCommit import CosiCommit
from .CosiCommitment import CosiCommitment
from .CosiSign import CosiSign
from .CosiSignature import CosiSignature
from .DebugLinkDecision import DebugLinkDecision
from .DebugLinkFlashErase import DebugLinkFlashErase
from .DebugLinkGetState import DebugLinkGetState
from .DebugLinkLog import DebugLinkLog
from .DebugLinkMemory import DebugLinkMemory
from .DebugLinkMemoryRead import DebugLinkMemoryRead
from .DebugLinkMemoryWrite import DebugLinkMemoryWrite
from .DebugLinkState import DebugLinkState
from .DebugLinkStop import DebugLinkStop
from .DecryptMessage import DecryptMessage
from .DecryptedMessage import DecryptedMessage
from .ECDHSessionKey import ECDHSessionKey
from .EncryptMessage import EncryptMessage
from .EncryptedMessage import EncryptedMessage
from .Entropy import Entropy
from .EntropyAck import EntropyAck
from .EntropyRequest import EntropyRequest
from .EstimateTxSize import EstimateTxSize
from .EthereumAddress import EthereumAddress
from .EthereumGetAddress import EthereumGetAddress
from .EthereumMessageSignature import EthereumMessageSignature
from .EthereumSignMessage import EthereumSignMessage
from .EthereumSignTx import EthereumSignTx
from .EthereumTxAck import EthereumTxAck
from .EthereumTxRequest import EthereumTxRequest
from .EthereumVerifyMessage import EthereumVerifyMessage
from .Failure import Failure
from .Features import Features
from .FirmwareErase import FirmwareErase
from .FirmwareRequest import FirmwareRequest
from .FirmwareUpload import FirmwareUpload
from .GetAddress import GetAddress
from .GetECDHSessionKey import GetECDHSessionKey
from .GetEntropy import GetEntropy
from .GetFeatures import GetFeatures
from .GetPublicKey import GetPublicKey
from .Initialize import Initialize
from .LiskAddress import LiskAddress
from .LiskGetAddress import LiskGetAddress
from .LiskGetPublicKey import LiskGetPublicKey
from .LiskMessageSignature import LiskMessageSignature
from .LiskPublicKey import LiskPublicKey
from .LiskSignMessage import LiskSignMessage
from .LiskSignTx import LiskSignTx
from .LiskSignedTx import LiskSignedTx
from .LiskVerifyMessage import LiskVerifyMessage
from .LoadDevice import LoadDevice
from .MessageSignature import MessageSignature
from .NEMAddress import NEMAddress
from .NEMDecryptMessage import NEMDecryptMessage
from .NEMDecryptedMessage import NEMDecryptedMessage
from .NEMGetAddress import NEMGetAddress
from .NEMSignTx import NEMSignTx
from .NEMSignedTx import NEMSignedTx
from .PassphraseAck import PassphraseAck
from .PassphraseRequest import PassphraseRequest
from .PassphraseStateAck import PassphraseStateAck
from .PassphraseStateRequest import PassphraseStateRequest
from .PinMatrixAck import PinMatrixAck
from .PinMatrixRequest import PinMatrixRequest
from .Ping import Ping
from .PublicKey import PublicKey
from .RecoveryDevice import RecoveryDevice
from .ResetDevice import ResetDevice
from .SelfTest import SelfTest
from .SetU2FCounter import SetU2FCounter
from .SignIdentity import SignIdentity
from .SignMessage import SignMessage
from .SignTx import SignTx
from .SignedIdentity import SignedIdentity
from .SimpleSignTx import SimpleSignTx
from .StellarAccountMergeOp import StellarAccountMergeOp
from .StellarAddress import StellarAddress
from .StellarAllowTrustOp import StellarAllowTrustOp
from .StellarBumpSequenceOp import StellarBumpSequenceOp
from .StellarChangeTrustOp import StellarChangeTrustOp
from .StellarCreateAccountOp import StellarCreateAccountOp
from .StellarCreatePassiveOfferOp import StellarCreatePassiveOfferOp
from .StellarGetAddress import StellarGetAddress
from .StellarGetPublicKey import StellarGetPublicKey
from .StellarManageDataOp import StellarManageDataOp
from .StellarManageOfferOp import StellarManageOfferOp
from .StellarPathPaymentOp import StellarPathPaymentOp
from .StellarPaymentOp import StellarPaymentOp
from .StellarPublicKey import StellarPublicKey
from .StellarSetOptionsOp import StellarSetOptionsOp
from .StellarSignTx import StellarSignTx
from .StellarSignedTx import StellarSignedTx
from .StellarTxOpRequest import StellarTxOpRequest
from .Success import Success
from .TxAck import TxAck
from .TxRequest import TxRequest
from .TxSize import TxSize
from .VerifyMessage import VerifyMessage
from .WipeDevice import WipeDevice
from .WordAck import WordAck
from .WordRequest import WordRequest
from . import MessageType



# Automatically generated by pb2py
from .. import protobuf as p


class ApplyFlags(p.MessageType):
    MESSAGE_WIRE_TYPE = 28
    FIELDS = {
        1: ('flags', p.UVarintType, 0),
    }

    def __init__(
        self,
        flags: int = None
    ) -> None:
        self.flags = flags



# Automatically generated by pb2py
from .. import protobuf as p
from .TxRequestDetailsType import TxRequestDetailsType
from .TxRequestSerializedType import TxRequestSerializedType


class TxRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 21
    FIELDS = {
        1: ('request_type', p.UVarintType, 0),
        2: ('details', TxRequestDetailsType, 0),
        3: ('serialized', TxRequestSerializedType, 0),
    }

    def __init__(
        self,
        request_type: int = None,
        details: TxRequestDetailsType = None,
        serialized: TxRequestSerializedType = None
    ) -> None:
        self.request_type = request_type
        self.details = details
        self.serialized = serialized



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import binascii
import hashlib
import pytest

from safewiselib import cosi

# These tests calculate Ed25519 signatures in pure Python.
# In addition to being Python, this is also DJB's proof-of-concept, unoptimized code.
# As a result, it is actually very noticeably slow. On a gen8 Core i5, this takes around 40 seconds.
# To skip the test, run `pytest -m "not slow_cosi"`.

# Therefore, the tests are skipped by default.
# Run `pytest -m slow_cosi` to explicitly enable.

pytestmark = pytest.mark.slow_cosi
if "slow_cosi" not in pytest.config.getoption("-m"):
    pytestmark = pytest.mark.skip("Skipping slow CoSi tests. 'pytest -m slow_cosi' to run.")


RFC8032_VECTORS = (
    (  # test 1
        # privkey
        binascii.unhexlify("9d61b19deffd5a60ba844af492ec2cc44449c5697b326919703bac031cae7f60"),
        # pubkey
        binascii.unhexlify("d75a980182b10ab7d54bfed3c964073a0ee172f3daa62325af021a68f707511a"),
        # message
        binascii.unhexlify(""),
        # signature
        binascii.unhexlify("e5564300c360ac729086e2cc806e828a84877f1eb8e5d974d873e06522490155"
                           "5fb8821590a33bacc61e39701cf9b46bd25bf5f0595bbe24655141438e7a100b"),
    ),
    (  # test 2
        # privkey
        binascii.unhexlify("4ccd089b28ff96da9db6c346ec114e0f5b8a319f35aba624da8cf6ed4fb8a6fb"),
        # pubkey
        binascii.unhexlify("3d4017c3e843895a92b70aa74d1b7ebc9c982ccf2ec4968cc0cd55f12af4660c"),
        # message
        binascii.unhexlify("72"),
        # signature
        binascii.unhexlify("92a009a9f0d4cab8720e820b5f642540a2b27b5416503f8fb3762223ebdb69da"
                           "085ac1e43e15996e458f3613d0f11d8c387b2eaeb4302aeeb00d291612bb0c00"),
    ),
    (  # test 3
        # privkey
        binascii.unhexlify("c5aa8df43f9f837bedb7442f31dcb7b166d38535076f094b85ce3a2e0b4458f7"),
        # pubkey
        binascii.unhexlify("fc51cd8e6218a1a38da47ed00230f0580816ed13ba3303ac5deb911548908025"),
        # message
        binascii.unhexlify("af82"),
        # signature
        binascii.unhexlify("6291d657deec24024827e69c3abe01a30ce548a284743a445e3680d7db5ac3ac"
                           "18ff9b538d16f290ae67f760984dc6594a7c15e9716ed28dc027beceea1ec40a"),
    ),
    (  # test SHA(abc)
        # privkey
        binascii.unhexlify("833fe62409237b9d62ec77587520911e9a759cec1d19755b7da901b96dca3d42"),
        # pubkey
        binascii.unhexlify("ec172b93ad5e563bf4932c70e1245034c35467ef2efd4d64ebf819683467e2bf"),
        # message
        hashlib.sha512(b"abc").digest(),
        # signature
        binascii.unhexlify("dc2a4459e7369633a52b1bf277839a00201009a3efbf3ecb69bea2186c26b589"
                           "09351fc9ac90b3ecfdfbc7c66431e0303dca179c138ac17ad9bef1177331a704"),
    ),
)

COMBINED_KEY = binascii.unhexlify("283967b1c19ff93d2924cdcba95e586547cafef509ea402963ceefe96ccb44f2")
GLOBAL_COMMIT = binascii.unhexlify("75bd5806c6366e0374a1c6e020c53feb0791d6cc07560d27d8c158f886ecf389")


@pytest.mark.parametrize("privkey, pubkey, message, signature",
                         RFC8032_VECTORS)
def test_single_eddsa_vector(privkey, pubkey, message, signature):
    my_pubkey = cosi.pubkey_from_privkey(privkey)
    assert my_pubkey == pubkey
    try:
        cosi.verify(signature, message, pubkey)
    except ValueError:
        pytest.fail("Signature does not verify.")

    fake_signature = b'\xf1' + signature[1:]
    with pytest.raises(ValueError):
        cosi.verify(fake_signature, message, pubkey)


def test_combine_keys():
    pubkeys = [pubkey for _, pubkey, _, _ in RFC8032_VECTORS]
    assert cosi.combine_keys(pubkeys) == COMBINED_KEY

    Rs = [cosi.get_nonce(privkey, message)[1] for privkey, _, message, _ in RFC8032_VECTORS]
    assert cosi.combine_keys(Rs) == GLOBAL_COMMIT


@pytest.mark.parametrize("keyset", [
    (0,),
    (0, 1),
    (0, 1, 2),
    (0, 1, 2, 3),
    (1, 3),
])
def test_cosi_combination(keyset):
    message = hashlib.sha512(b"You all have to sign this.").digest()
    selection = [RFC8032_VECTORS[i] for i in keyset]

    # zip(*iterable) turns a list of tuples to a tuple of lists
    privkeys, pubkeys, _, _ = zip(*selection)
    nonce_pairs = [cosi.get_nonce(pk, message) for pk in privkeys]
    nonces, commits = zip(*nonce_pairs)

    # calculate global pubkey and commitment
    global_pk = cosi.combine_keys(pubkeys)
    global_commit = cosi.combine_keys(commits)

    # generate individual signatures
    signatures = [cosi.sign_with_privkey(message, privkey, global_pk, nonce, global_commit)
                  for privkey, nonce in zip(privkeys, nonces)]

    # combine signatures
    global_sig = cosi.combine_sig(global_commit, signatures)

    try:
        cosi.verify(global_sig, message, global_pk)
    except ValueError:
        pytest.fail("Failed to validate global signature")



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import warnings

from .transport import enumerate_devices, get_transport


class SafeWISEDevice:
    '''
    This class is deprecated. (There is no reason for it to exist in the first
    place, it is nothing but a collection of two functions.)
    Instead, please use functions from the ``safewiselib.transport`` module.
    '''

    @classmethod
    def enumerate(cls):
        warnings.warn('SafeWISEDevice is deprecated.', DeprecationWarning)
        return enumerate_devices()

    @classmethod
    def find_by_path(cls, path):
        warnings.warn('SafeWISEDevice is deprecated.', DeprecationWarning)
        return get_transport(path, prefix_search=False)



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import os.path
import json

from .tx_api import TxApiInsight, TxApiBlockCypher

COINS_JSON = os.path.join(os.path.dirname(__file__), 'coins.json')


def _load_coins_json():
    # Load coins.json to local variables
    # NOTE: coins.json comes from 'vendor/safewise-common/coins.json',
    # which is a git submodule. If you're trying to run safewiselib directly
    # from the checkout (or tarball), initialize the submodule with:
    # $ git submodule update --init
    # and install coins.json with:
    # $ python setup.py prebuild
    with open(COINS_JSON) as coins_json:
        return json.load(coins_json)


def _insight_for_coin(coin):
    if not coin['bitcore']:
        return None
    zcash = coin['coin_name'].lower().startswith('zcash')
    network = 'insight_{}'.format(coin['coin_name'].lower().replace(' ', '_'))
    url = coin['bitcore'][0] + '/api/'
    return TxApiInsight(network=network, url=url, zcash=zcash)


# exported variables
__all__ = ['by_name', 'slip44', 'tx_api']

try:
    by_name = _load_coins_json()
except Exception as e:
    raise ImportError("Failed to load coins.json. Check your installation.") from e

slip44 = {name: coin['slip44'] for name, coin in by_name.items()}
tx_api = {name: _insight_for_coin(coin)
          for name, coin in by_name.items()
          if coin["bitcore"]}

# fixup for Dogecoin
tx_api['Dogecoin'] = TxApiBlockCypher(network='blockcypher_dogecoin', url='https://api.blockcypher.com/v1/doge/main/')



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from binascii import hexlify
import pytest
import time

from .common import SafeWISETest
from safewiselib import messages as proto
from safewiselib import nem
from safewiselib.tools import parse_path


# assertion data from T1
@pytest.mark.nem
@pytest.mark.skip_t1
class TestMsgNEMSignTxMosaics(SafeWISETest):

    def test_nem_signtx_mosaic_supply_change(self):
        self.setup_mnemonic_nopin_nopassphrase()

        with self.client:
            tx = self.client.nem_sign_tx(parse_path("m/44'/1'/0'/0'/0'"), {
                "timeStamp": 74649215,
                "fee": 2000000,
                "type": nem.TYPE_MOSAIC_SUPPLY_CHANGE,
                "deadline": 74735615,
                "message": {
                },
                "mosaicId": {
                    "namespaceId": "hellom",
                    "name": "Hello mosaic"
                },
                "supplyType": 1,
                "delta": 1,
                "version": (0x98 << 24),
                "creationFeeSink": "TALICE2GMA34CXHD7XLJQ536NM5UNKQHTORNNT2J",
                "creationFee": 1500,
            })

            assert hexlify(tx.data) == b'02400000010000987f0e730420000000edfd32f6e760648c032f9acb4b30d514265f6a5b5f8a7154f2618922b406208480841e0000000000ff5f74041a0000000600000068656c6c6f6d0c00000048656c6c6f206d6f73616963010000000100000000000000'
            assert hexlify(tx.signature) == b'928b03c4a69fff35ecf0912066ea705895b3028fad141197d7ea2b56f1eef2a2516455e6f35d318f6fa39e2bb40492ac4ae603260790f7ebc7ea69feb4ca4c0a'

    def test_nem_signtx_mosaic_creation(self):
        self.setup_mnemonic_nopin_nopassphrase()

        test_suite = {
            "timeStamp": 74649215,
            "fee": 2000000,
            "type": nem.TYPE_MOSAIC_CREATION,
            "deadline": 74735615,
            "message": {
            },
            "mosaicDefinition": {
                "id": {
                    "namespaceId": "hellom",
                    "name": "Hello mosaic"
                },
                "levy": {},
                "properties": {},
                "description": "lorem"
            },
            "version": (0x98 << 24),
            "creationFeeSink": "TALICE2GMA34CXHD7XLJQ536NM5UNKQHTORNNT2J",
            "creationFee": 1500,
        }

        # not using client.nem_sign_tx() because of swiping
        tx = self._nem_sign(2, test_suite)
        assert hexlify(tx.data) == b'01400000010000987f0e730420000000edfd32f6e760648c032f9acb4b30d514265f6a5b5f8a7154f2618922b406208480841e0000000000ff5f7404c100000020000000edfd32f6e760648c032f9acb4b30d514265f6a5b5f8a7154f2618922b40620841a0000000600000068656c6c6f6d0c00000048656c6c6f206d6f73616963050000006c6f72656d04000000150000000c00000064697669736962696c6974790100000030160000000d000000696e697469616c537570706c7901000000301a0000000d000000737570706c794d757461626c650500000066616c7365190000000c0000007472616e7366657261626c650500000066616c7365000000002800000054414c49434532474d4133344358484437584c4a513533364e4d35554e4b5148544f524e4e54324adc05000000000000'
        assert hexlify(tx.signature) == b'537adf4fd9bd5b46e204b2db0a435257a951ed26008305e0aa9e1201dafa4c306d7601a8dbacabf36b5137724386124958d53202015ab31fb3d0849dfed2df0e'

    def test_nem_signtx_mosaic_creation_properties(self):
        self.setup_mnemonic_nopin_nopassphrase()

        test_suite = {
            "timeStamp": 74649215,
            "fee": 2000000,
            "type": nem.TYPE_MOSAIC_CREATION,
            "deadline": 74735615,
            "message": {
            },
            "mosaicDefinition": {
                "id": {
                    "namespaceId": "hellom",
                    "name": "Hello mosaic"
                },
                "levy": {},
                "properties": [
                    {
                        "name": "divisibility",
                        "value": "4"
                    },
                    {
                        "name": "initialSupply",
                        "value": "200"
                    },
                    {
                        "name": "supplyMutable",
                        "value": "false"
                    },
                    {
                        "name": "transferable",
                        "value": "true"
                    }
                ],
                "description": "lorem"
            },
            "version": (0x98 << 24),
            "creationFeeSink": "TALICE2GMA34CXHD7XLJQ536NM5UNKQHTORNNT2J",
            "creationFee": 1500,
        }

        # not using client.nem_sign_tx() because of swiping
        tx = self._nem_sign(2, test_suite)
        assert hexlify(tx.data) == b'01400000010000987f0e730420000000edfd32f6e760648c032f9acb4b30d514265f6a5b5f8a7154f2618922b406208480841e0000000000ff5f7404c200000020000000edfd32f6e760648c032f9acb4b30d514265f6a5b5f8a7154f2618922b40620841a0000000600000068656c6c6f6d0c00000048656c6c6f206d6f73616963050000006c6f72656d04000000150000000c00000064697669736962696c6974790100000034180000000d000000696e697469616c537570706c79030000003230301a0000000d000000737570706c794d757461626c650500000066616c7365180000000c0000007472616e7366657261626c650400000074727565000000002800000054414c49434532474d4133344358484437584c4a513533364e4d35554e4b5148544f524e4e54324adc05000000000000'
        assert hexlify(tx.signature) == b'f17c859710060f2ea9a0ab740ef427431cf36bdc7d263570ca282bd66032e9f5737a921be9839429732e663be2bb74ccc16f34f5157ff2ef00a65796b54e800e'

    def test_nem_signtx_mosaic_creation_levy(self):
        self.setup_mnemonic_nopin_nopassphrase()

        test_suite = {
            "timeStamp": 74649215,
            "fee": 2000000,
            "type": nem.TYPE_MOSAIC_CREATION,
            "deadline": 74735615,
            "message": {
            },
            "mosaicDefinition": {
                "id": {
                    "namespaceId": "hellom",
                    "name": "Hello mosaic"
                },
                "properties": [
                    {
                        "name": "divisibility",
                        "value": "4"
                    },
                    {
                        "name": "initialSupply",
                        "value": "200"
                    },
                    {
                        "name": "supplyMutable",
                        "value": "false"
                    },
                    {
                        "name": "transferable",
                        "value": "true"
                    }
                ],
                "levy": {
                    "type": 1,
                    "fee": 2,
                    "recipient": "TALICE2GMA34CXHD7XLJQ536NM5UNKQHTORNNT2J",
                    "mosaicId": {
                        "namespaceId": "hellom",
                        "name": "Hello mosaic"
                    },
                },
                "description": "lorem"
            },
            "version": (0x98 << 24),
            "creationFeeSink": "TALICE2GMA34CXHD7XLJQ536NM5UNKQHTORNNT2J",
            "creationFee": 1500,
        }

        tx = self._nem_sign(6, test_suite)
        assert hexlify(tx.data) == b'01400000010000987f0e730420000000edfd32f6e760648c032f9acb4b30d514265f6a5b5f8a7154f2618922b406208480841e0000000000ff5f74041801000020000000edfd32f6e760648c032f9acb4b30d514265f6a5b5f8a7154f2618922b40620841a0000000600000068656c6c6f6d0c00000048656c6c6f206d6f73616963050000006c6f72656d04000000150000000c00000064697669736962696c6974790100000034180000000d000000696e697469616c537570706c79030000003230301a0000000d000000737570706c794d757461626c650500000066616c7365180000000c0000007472616e7366657261626c65040000007472756556000000010000002800000054414c49434532474d4133344358484437584c4a513533364e4d35554e4b5148544f524e4e54324a1a0000000600000068656c6c6f6d0c00000048656c6c6f206d6f7361696302000000000000002800000054414c49434532474d4133344358484437584c4a513533364e4d35554e4b5148544f524e4e54324adc05000000000000'
        assert hexlify(tx.signature) == b'b87aac1ddf146d35e6a7f3451f57e2fe504ac559031e010a51261257c37bd50fcfa7b2939dd7a3203b54c4807d458475182f5d3dc135ec0d1d4a9cd42159fd0a'

    def _nem_sign(self, num_of_swipes, test_suite):
        n = parse_path("m/44'/1'/0'/0'/0'")
        n = self.client._convert_prime(n)
        msg = nem.create_sign_tx(test_suite)
        assert msg.transaction is not None
        msg.transaction.address_n = n

        # Sending NEMSignTx message
        self.client.transport.write(msg)
        ret = self.client.transport.read()

        # Confirm action
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        self.client.transport.write(proto.ButtonAck())
        time.sleep(1)
        for i in range(num_of_swipes):
            self.client.debug.swipe_down()
            time.sleep(1)
        self.client.debug.press_yes()
        ret = self.client.transport.read()

        # Confirm action
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        self.client.transport.write(proto.ButtonAck())
        ret = self.client.transport.read()

        # Confirm tx
        assert isinstance(ret, proto.ButtonRequest)
        self.client.debug.press_yes()
        self.client.transport.write(proto.ButtonAck())
        return self.client.transport.read()



# Automatically generated by pb2py
from .. import protobuf as p


class CosiCommitment(p.MessageType):
    MESSAGE_WIRE_TYPE = 72
    FIELDS = {
        1: ('commitment', p.BytesType, 0),
        2: ('pubkey', p.BytesType, 0),
    }

    def __init__(
        self,
        commitment: bytes = None,
        pubkey: bytes = None
    ) -> None:
        self.commitment = commitment
        self.pubkey = pubkey



# Automatically generated by pb2py
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None


class EncryptMessage(p.MessageType):
    MESSAGE_WIRE_TYPE = 49
    FIELDS = {
        1: ('pubkey', p.BytesType, 0),
        2: ('message', p.BytesType, 0),
        3: ('display_only', p.BoolType, 0),
        4: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        5: ('coin_name', p.UnicodeType, 0),  # default='Bitcoin'
    }

    def __init__(
        self,
        pubkey: bytes = None,
        message: bytes = None,
        display_only: bool = None,
        address_n: List[int] = None,
        coin_name: str = None
    ) -> None:
        self.pubkey = pubkey
        self.message = message
        self.display_only = display_only
        self.address_n = address_n if address_n is not None else []
        self.coin_name = coin_name



# Automatically generated by pb2py
from .. import protobuf as p


class GetEntropy(p.MessageType):
    MESSAGE_WIRE_TYPE = 9
    FIELDS = {
        1: ('size', p.UVarintType, 0),  # required
    }

    def __init__(
        self,
        size: int = None
    ) -> None:
        self.size = size



# Automatically generated by pb2py
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None
from .MultisigRedeemScriptType import MultisigRedeemScriptType


class TxInputType(p.MessageType):
    FIELDS = {
        1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        2: ('prev_hash', p.BytesType, 0),  # required
        3: ('prev_index', p.UVarintType, 0),  # required
        4: ('script_sig', p.BytesType, 0),
        5: ('sequence', p.UVarintType, 0),  # default=4294967295
        6: ('script_type', p.UVarintType, 0),  # default=0
        7: ('multisig', MultisigRedeemScriptType, 0),
        8: ('amount', p.UVarintType, 0),
        9: ('decred_tree', p.UVarintType, 0),
        10: ('decred_script_version', p.UVarintType, 0),
    }

    def __init__(
        self,
        address_n: List[int] = None,
        prev_hash: bytes = None,
        prev_index: int = None,
        script_sig: bytes = None,
        sequence: int = None,
        script_type: int = None,
        multisig: MultisigRedeemScriptType = None,
        amount: int = None,
        decred_tree: int = None,
        decred_script_version: int = None
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.prev_hash = prev_hash
        self.prev_index = prev_index
        self.script_sig = script_sig
        self.sequence = sequence
        self.script_type = script_type
        self.multisig = multisig
        self.amount = amount
        self.decred_tree = decred_tree
        self.decred_script_version = decred_script_version



#!/usr/bin/env python3
from safewiselib.debuglink import DebugLink
from safewiselib.client import SafeWISEClient
from safewiselib.transport import enumerate_devices
import binascii
import sys


def main():
    # List all debuggable SafeWISEs
    devices = [device for device in enumerate_devices() if hasattr(device, 'find_debug')]

    # Check whether we found any
    if len(devices) == 0:
        print('No SafeWISE found')
        return

    # Use first connected device
    transport = devices[0]
    debug_transport = devices[0].find_debug()

    # Creates object for manipulating SafeWISE
    client = SafeWISEClient(transport)
    debug = DebugLink(debug_transport)

    debug.memory_write(int(sys.argv[1], 16), binascii.unhexlify(sys.argv[2]), flash=True)
    client.close()


if __name__ == '__main__':
    main()



# Automatically generated by pb2py
from safewiselib import protobuf as p


class CoinDef(p.MessageType):
    FIELDS = {
        1: ('coin_name', p.UnicodeType, 0),
        2: ('coin_shortcut', p.UnicodeType, 0),
        3: ('coin_label', p.UnicodeType, 0),
        4: ('curve_name', p.UnicodeType, 0),
        5: ('address_type', p.UVarintType, 0),
        6: ('address_type_p2sh', p.UVarintType, 0),
        7: ('maxfee_kb', p.UVarintType, 0),
        8: ('minfee_kb', p.UVarintType, 0),
        9: ('signed_message_header', p.BytesType, 0),
        10: ('hash_genesis_block', p.BytesType, 0),
        11: ('xprv_magic', p.UVarintType, 0),
        12: ('xpub_magic', p.UVarintType, 0),
        13: ('xpub_magic_segwit_p2sh', p.UVarintType, 0),
        14: ('xpub_magic_segwit_native', p.UVarintType, 0),
        15: ('bech32_prefix', p.UnicodeType, 0),
        16: ('cashaddr_prefix', p.UnicodeType, 0),
        17: ('slip44', p.UVarintType, 0),
        18: ('segwit', p.BoolType, 0),
        19: ('decred', p.BoolType, 0),
        20: ('fork_id', p.UVarintType, 0),
        21: ('force_bip143', p.BoolType, 0),
        22: ('dust_limit', p.UVarintType, 0),
        23: ('address_prefix', p.UnicodeType, 0),
        24: ('min_address_length', p.UVarintType, 0),
        25: ('max_address_length', p.UVarintType, 0),
        26: ('icon', p.BytesType, 0),
        27: ('version_group_id', p.UVarintType, 0),
        28: ('website', p.UnicodeType, 0),
        29: ('github', p.UnicodeType, 0),
        30: ('maintainer', p.UnicodeType, 0),
        31: ('blocktime_seconds', p.UVarintType, 0),
    }

    def __init__(
        self,
        coin_name: str = None,
        coin_shortcut: str = None,
        coin_label: str = None,
        curve_name: str = None,
        address_type: int = None,
        address_type_p2sh: int = None,
        maxfee_kb: int = None,
        minfee_kb: int = None,
        signed_message_header: bytes = None,
        hash_genesis_block: bytes = None,
        xprv_magic: int = None,
        xpub_magic: int = None,
        xpub_magic_segwit_p2sh: int = None,
        xpub_magic_segwit_native: int = None,
        bech32_prefix: str = None,
        cashaddr_prefix: str = None,
        slip44: int = None,
        segwit: bool = None,
        decred: bool = None,
        fork_id: int = None,
        force_bip143: bool = None,
        dust_limit: int = None,
        address_prefix: str = None,
        min_address_length: int = None,
        max_address_length: int = None,
        icon: bytes = None,
        version_group_id: int = None,
        website: str = None,
        github: str = None,
        maintainer: str = None,
        blocktime_seconds: int = None,
        default_fee_b: dict = None,
        bitcore: dict = None,
        blockbook: dict = None
    ):
        self.coin_name = coin_name
        self.coin_shortcut = coin_shortcut
        self.coin_label = coin_label
        self.curve_name = curve_name
        self.address_type = address_type
        self.address_type_p2sh = address_type_p2sh
        self.maxfee_kb = maxfee_kb
        self.minfee_kb = minfee_kb
        self.signed_message_header = signed_message_header
        self.hash_genesis_block = hash_genesis_block
        self.xprv_magic = xprv_magic
        self.xpub_magic = xpub_magic
        self.xpub_magic_segwit_p2sh = xpub_magic_segwit_p2sh
        self.xpub_magic_segwit_native = xpub_magic_segwit_native
        self.bech32_prefix = bech32_prefix
        self.cashaddr_prefix = cashaddr_prefix
        self.slip44 = slip44
        self.segwit = segwit
        self.decred = decred
        self.fork_id = fork_id
        self.force_bip143 = force_bip143
        self.dust_limit = dust_limit
        self.address_prefix = address_prefix
        self.min_address_length = min_address_length
        self.max_address_length = max_address_length
        self.icon = icon
        self.version_group_id = version_group_id
        self.website = website
        self.github = github
        self.maintainer = maintainer
        self.blocktime_seconds = blocktime_seconds
        self.default_fee_b = default_fee_b
        self.bitcore = bitcore
        self.blockbook = blockbook
        p.MessageType.__init__(self)



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest
from binascii import hexlify, unhexlify

from .common import SafeWISETest
from ..support.ckd_public import deserialize
from safewiselib import coins
from safewiselib import messages as proto
from safewiselib.client import CallException
from safewiselib.tools import parse_path

TxApiBitcoinGold = coins.tx_api["Bgold"]


# All data taken from T1
class TestMsgSigntxBitcoinGold(SafeWISETest):

    def test_send_bitcoin_gold_change(self):
        self.setup_mnemonic_allallall()
        self.client.set_tx_api(TxApiBitcoinGold)
        inp1 = proto.TxInputType(
            address_n=parse_path("44'/156'/0'/0/0"),
            amount=1995344,
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=0,
            script_type=proto.InputScriptType.SPENDADDRESS,
        )
        out1 = proto.TxOutputType(
            address_n=parse_path("44'/156'/0'/1/0"),
            amount=1896050,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )
        out2 = proto.TxOutputType(
            address='GfDB1tvjfm3bukeoBTtfNqrJVFohS2kCTe',
            amount=73452,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )
        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures, serialized_tx) = self.client.sign_tx('Bgold', [inp1], [out1, out2])

        assert hexlify(serialized_tx) == b'010000000185c9dd4ae1071affd77d90b9d03c1b5fdd7c62cf30a9bb8230ad766cf06b5225000000006b483045022100963904da0731b71ce468afd45366dd80fbff566ec0d39c1161ab85d17459c7ca02202f5c24a7a7272d98b14a3f5bc000c7cde8ac0eb773f20f4c3131518186cc98854121023bd0ec4022d12d0106c5b7308a25572953ba1951f576f691354a7b147ee0cc1fffffffff0272ee1c00000000001976a9141c82b9c11f193ad82413caadc0955730572b50ae88acec1e0100000000001976a914ea5f904d195079a350b534db4446433b3cec222e88ac00000000'

    def test_send_bitcoin_gold_nochange(self):
        self.setup_mnemonic_allallall()
        self.client.set_tx_api(TxApiBitcoinGold)
        inp1 = proto.TxInputType(
            address_n=parse_path("44'/156'/0'/1/0"),
            amount=1896050,
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=0,
            script_type=proto.InputScriptType.SPENDADDRESS,
        )
        inp2 = proto.TxInputType(
            address_n=parse_path("44'/156'/0'/0/1"),
            # 1LRspCZNFJcbuNKQkXgHMDucctFRQya5a3
            amount=73452,
            prev_hash=unhexlify('db77c2461b840e6edbe7f9280043184a98e020d9795c1b65cb7cef2551a8fb18'),
            prev_index=1,
            script_type=proto.InputScriptType.SPENDADDRESS,
        )
        out1 = proto.TxOutputType(
            address='GfDB1tvjfm3bukeoBTtfNqrJVFohS2kCTe',
            amount=1934960,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )
        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures, serialized_tx) = self.client.sign_tx('Bgold', [inp1, inp2], [out1])

        assert hexlify(serialized_tx) == b'010000000285c9dd4ae1071affd77d90b9d03c1b5fdd7c62cf30a9bb8230ad766cf06b5225000000006b483045022100928852076c9fab160c07564cd54691af1cbc37fb28f0b7bee7299c7925ef62f0022058856387afecc6508f2f04ecdfd292a13026a5b2107ebdd2cc789bdf8820d552412102a6c3998d0d4e5197ff41aab5c53580253b3b91f583f4c31f7624be7dc83ce15fffffffff18fba85125ef7ccb651b5c79d920e0984a18430028f9e7db6e0e841b46c277db010000006b483045022100faa2f4f01cc95e680349a093923aae0aa2ea01429873555aa8a84bf630ef33a002204c3f4bf567e2d20540c0f71dc278481d6ccb6b95acda2a2f87ce521c79d6b872412102d54a7e5733b1635e5e9442943f48179b1700206b2d1925250ba10f1c86878be8ffffffff0170861d00000000001976a914ea5f904d195079a350b534db4446433b3cec222e88ac00000000'

    def test_attack_change_input(self):
        self.setup_mnemonic_allallall()
        self.client.set_tx_api(TxApiBitcoinGold)
        inp1 = proto.TxInputType(
            address_n=parse_path("44'/156'/1000'/0/0"),
            # 1MH9KKcvdCTY44xVDC2k3fjBbX5Cz29N1q
            amount=1995344,
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=0,
            script_type=proto.InputScriptType.SPENDADDRESS,
        )
        out1 = proto.TxOutputType(
            address_n=parse_path("44'/156'/1000'/1/0"),
            amount=1896050,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )
        out2 = proto.TxOutputType(
            address='GfDB1tvjfm3bukeoBTtfNqrJVFohS2kCTe',
            amount=73452,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )

        global attack_ctr
        attack_ctr = 0

        def attack_processor(req, msg):
            global attack_ctr

            if req.details.tx_hash is not None:
                return msg

            if req.request_type != proto.RequestType.TXINPUT:
                return msg

            attack_ctr += 1
            if attack_ctr <= 1:
                return msg

            msg.inputs[0].address_n[2] = 1 + 0x80000000
            return msg

        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.Failure(code=proto.FailureType.ProcessError),
            ])
            with pytest.raises(CallException):
                self.client.sign_tx('Bgold', [inp1], [out1, out2], debug_processor=attack_processor)

    def test_send_bch_multisig_change(self):
        self.setup_mnemonic_allallall()
        self.client.set_tx_api(TxApiBitcoinGold)
        xpubs = []
        for n in map(lambda index: self.client.get_public_node(parse_path("44'/156'/" + str(index) + "'")), range(1, 4)):
            xpubs.append(n.xpub)

        def getmultisig(chain, nr, signatures=[b'', b'', b''], xpubs=xpubs):
            return proto.MultisigRedeemScriptType(
                pubkeys=list(map(lambda xpub: proto.HDNodePathType(node=deserialize(xpub), address_n=[chain, nr]), xpubs)),
                signatures=signatures,
                m=2,
            )
        inp1 = proto.TxInputType(
            address_n=parse_path("44'/156'/3'/0/0"),
            multisig=getmultisig(0, 0),
            # 33Ju286QvonBz5N1V754ZekQv4GLJqcc5R
            amount=48490,
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=0,
            script_type=proto.InputScriptType.SPENDMULTISIG,
        )
        out1 = proto.TxOutputType(
            address='GfDB1tvjfm3bukeoBTtfNqrJVFohS2kCTe',
            amount=24000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )
        out2 = proto.TxOutputType(
            address_n=parse_path("44'/156'/3'/1/0"),
            multisig=getmultisig(1, 0),
            script_type=proto.OutputScriptType.PAYTOMULTISIG,
            amount=24000
        )
        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures1, serialized_tx) = self.client.sign_tx('Bgold', [inp1], [out1, out2])

        assert hexlify(signatures1[0]) == b'3045022100b1594f3b186d0dedbf61e53a1c407b1e0747098b7375941df85af045040f578e022013ba1893eb9e2fd854dd07073a83b261cf4beba76f66b07742e462b4088a7e4a'

        inp1 = proto.TxInputType(
            address_n=parse_path("44'/156'/1'/0/0"),
            multisig=getmultisig(0, 0, [b'', b'', signatures1[0]]),
            # 33Ju286QvonBz5N1V754ZekQv4GLJqcc5R
            amount=48490,
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=0,
            script_type=proto.InputScriptType.SPENDMULTISIG,
        )
        out2.address_n[2] = 1 + 0x80000000

        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures1, serialized_tx) = self.client.sign_tx('Bgold', [inp1], [out1, out2])

        assert hexlify(signatures1[0]) == b'3044022006da8dbd14e6656ac8dcb956f4c0498574e88680eaeceb2cbafd8d2b2329d8cc02200972d076d444c5ff8f2ab18e14d8249ab661cb9c53335039bedcde037a40d747'
        assert hexlify(serialized_tx) == b'010000000185c9dd4ae1071affd77d90b9d03c1b5fdd7c62cf30a9bb8230ad766cf06b522500000000fdfd0000473044022006da8dbd14e6656ac8dcb956f4c0498574e88680eaeceb2cbafd8d2b2329d8cc02200972d076d444c5ff8f2ab18e14d8249ab661cb9c53335039bedcde037a40d74741483045022100b1594f3b186d0dedbf61e53a1c407b1e0747098b7375941df85af045040f578e022013ba1893eb9e2fd854dd07073a83b261cf4beba76f66b07742e462b4088a7e4a414c69522102290e6649574d17938c1ecb959ae92954f9ee48e1bd5b73f35ea931a3ab8a6087210379e0107b173e2c143426760627128c5eea3f862e8df92f3c2558eeeae4e347842103ff1746ca7dcf9e5c2eea9a73779b7c5bafed549f45cf3638a94cdf1e89c7f28f53aeffffffff02c05d0000000000001976a914ea5f904d195079a350b534db4446433b3cec222e88acc05d00000000000017a91445e917e46815d2b38d3f1cf072e63dd4f3b7a7e38700000000'

    def test_send_p2sh(self):
        self.setup_mnemonic_allallall()
        self.client.set_tx_api(TxApiBitcoinGold)
        inp1 = proto.TxInputType(
            address_n=parse_path("49'/156'/0'/1/0"),
            amount=123456789,
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=0,
            script_type=proto.InputScriptType.SPENDP2SHWITNESS,
        )
        out1 = proto.TxOutputType(
            address='GfDB1tvjfm3bukeoBTtfNqrJVFohS2kCTe',
            amount=12300000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )
        out2 = proto.TxOutputType(
            address='GZFLExxrvWFuFT1xRzhfwQWSE2bPDedBfn',
            script_type=proto.OutputScriptType.PAYTOADDRESS,
            amount=123456789 - 11000 - 12300000,
        )
        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures, serialized_tx) = self.client.sign_tx('Bgold', [inp1], [out1, out2])

        assert hexlify(serialized_tx) == b'0100000000010185c9dd4ae1071affd77d90b9d03c1b5fdd7c62cf30a9bb8230ad766cf06b52250000000017160014b5355d001e720d8f4513da00ff2bba4dcf9d39fcffffffff02e0aebb00000000001976a914ea5f904d195079a350b534db4446433b3cec222e88ac3df39f06000000001976a914a8f757819ec6779409f45788f7b4a0e8f51ec50488ac02473044022073fcbf2876f073f78923ab427f14de5b2a0fbeb313a9b2b650b3567061f242a702202f45fc22c501108ff6222afe3aca7da9d8c7dc860f9cda335bef31fa184e7bef412102ecea08b559fc5abd009acf77cfae13fa8a3b1933e3e031956c65c12cec8ca3e300000000'

    def test_send_p2sh_witness_change(self):
        self.setup_mnemonic_allallall()
        self.client.set_tx_api(TxApiBitcoinGold)
        inp1 = proto.TxInputType(
            address_n=parse_path("49'/156'/0'/1/0"),
            amount=123456789,
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=0,
            script_type=proto.InputScriptType.SPENDP2SHWITNESS,
        )
        out1 = proto.TxOutputType(
            address='GfDB1tvjfm3bukeoBTtfNqrJVFohS2kCTe',
            amount=12300000,
            script_type=proto.OutputScriptType.PAYTOADDRESS,
        )
        out2 = proto.TxOutputType(
            address_n=parse_path("49'/156'/0'/1/0"),
            script_type=proto.OutputScriptType.PAYTOP2SHWITNESS,
            amount=123456789 - 11000 - 12300000,
        )
        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=1)),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures, serialized_tx) = self.client.sign_tx('Bgold', [inp1], [out1, out2])

        # print(hexlify(serialized_tx))
        # assert False
        assert hexlify(serialized_tx) == b'0100000000010185c9dd4ae1071affd77d90b9d03c1b5fdd7c62cf30a9bb8230ad766cf06b52250000000017160014b5355d001e720d8f4513da00ff2bba4dcf9d39fcffffffff02e0aebb00000000001976a914ea5f904d195079a350b534db4446433b3cec222e88ac3df39f060000000017a9140cd03822b799a452c106d1b3771844a067b17f118702483045022100d79b33384c686d8dd40ad5f84f46691d30994992c1cb42e934c2a625d86cb2f902206859805a9a98ba140b71a9d4b9a6b8df94a9424f9c40f3bd804149fd6e278d63412102ecea08b559fc5abd009acf77cfae13fa8a3b1933e3e031956c65c12cec8ca3e300000000'

    def test_send_multisig_1(self):
        self.setup_mnemonic_allallall()
        self.client.set_tx_api(TxApiBitcoinGold)
        nodes = map(lambda index: self.client.get_public_node(parse_path("999'/1'/%d'" % index)), range(1, 4))
        multisig = proto.MultisigRedeemScriptType(
            pubkeys=list(map(lambda n: proto.HDNodePathType(node=deserialize(n.xpub), address_n=[2, 0]), nodes)),
            signatures=[b'', b'', b''],
            m=2,
        )

        inp1 = proto.TxInputType(
            address_n=parse_path("999'/1'/1'/2/0"),
            prev_hash=unhexlify('25526bf06c76ad3082bba930cf627cdd5f1b3cd0b9907dd7ff1a07e14addc985'),
            prev_index=1,
            script_type=proto.InputScriptType.SPENDP2SHWITNESS,
            multisig=multisig,
            amount=1610436
        )

        out1 = proto.TxOutputType(address='GfDB1tvjfm3bukeoBTtfNqrJVFohS2kCTe',
                                  amount=1605000,
                                  script_type=proto.OutputScriptType.PAYTOADDRESS)

        with self.client:
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures1, _) = self.client.sign_tx('Bgold', [inp1], [out1])
            # store signature
            inp1.multisig.signatures[0] = signatures1[0]
            # sign with third key
            inp1.address_n[2] = 0x80000003
            self.client.set_expected_responses([
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.ButtonRequest(code=proto.ButtonRequestType.ConfirmOutput),
                proto.ButtonRequest(code=proto.ButtonRequestType.SignTx),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXOUTPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXINPUT, details=proto.TxRequestDetailsType(request_index=0)),
                proto.TxRequest(request_type=proto.RequestType.TXFINISHED),
            ])
            (signatures2, serialized_tx) = self.client.sign_tx('Bgold', [inp1], [out1])

        assert hexlify(serialized_tx) == b'0100000000010185c9dd4ae1071affd77d90b9d03c1b5fdd7c62cf30a9bb8230ad766cf06b522501000000232200201e8dda334f11171190b3da72e526d441491464769679a319a2f011da5ad312a1ffffffff01887d1800000000001976a914ea5f904d195079a350b534db4446433b3cec222e88ac0400483045022100e728485c8337f9a09ebbf36edc0fef10f8bcf5c1ba601b7d8ba43a9250a898f002206b9e3401c297f9ab9afb7f1be59bb342db53b5b65aff7c557e3109679697df0f41473044022062ea69ecdc07d0dadc1971fbda50a629a56dd30f431db26327428f4992601ce602204a1c8ab9c7d81c36cb6f819109a26f9baaa9607b8d37bff5e24eee6fab4a04e441695221038e81669c085a5846e68e03875113ddb339ecbb7cb11376d4163bca5dc2e2a0c1210348c5c3be9f0e6cf1954ded1c0475beccc4d26aaa9d0cce2dd902538ff1018a112103931140ebe0fbbb7df0be04ed032a54e9589e30339ba7bbb8b0b71b15df1294da53ae00000000'



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import pytest

from .common import SafeWISETest


@pytest.mark.skip_t2
class TestDeviceLoadXprv(SafeWISETest):

    def test_load_device_xprv_1(self):
        self.client.load_device_by_xprv(xprv='xprv9s21ZrQH143K2JF8RafpqtKiTbsbaxEeUaMnNHsm5o6wCW3z8ySyH4UxFVSfZ8n7ESu7fgir8imbZKLYVBxFPND1pniTZ81vKfd45EHKX73', pin='', passphrase_protection=False, label='test', language='english')

        passphrase_protection = self.client.debug.read_passphrase_protection()
        assert passphrase_protection is False

        address = self.client.get_address('Bitcoin', [])
        assert address == '128RdrAkJDmqasgvfRf6MC5VcX4HKqH4mR'

    def test_load_device_xprv_2(self):
        self.client.load_device_by_xprv(xprv='xprv9s21ZrQH143K2JF8RafpqtKiTbsbaxEeUaMnNHsm5o6wCW3z8ySyH4UxFVSfZ8n7ESu7fgir8imbZKLYVBxFPND1pniTZ81vKfd45EHKX73', pin='', passphrase_protection=True, label='test', language='english')

        self.client.set_passphrase('passphrase')

        passphrase_protection = self.client.debug.read_passphrase_protection()
        assert passphrase_protection is True

        address = self.client.get_address('Bitcoin', [])
        assert address == '1CHUbFa4wTTPYgkYaw2LHSd5D4qJjMU8ri'



# Automatically generated by pb2py
from .. import protobuf as p


class StellarAccountMergeOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 218
    FIELDS = {
        1: ('source_account', p.BytesType, 0),
        2: ('destination_account', p.BytesType, 0),
    }

    def __init__(
        self,
        source_account: bytes = None,
        destination_account: bytes = None
    ) -> None:
        self.source_account = source_account
        self.destination_account = destination_account



# Automatically generated by pb2py
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None
from .TransactionType import TransactionType
from .TxInputType import TxInputType
from .TxOutputType import TxOutputType


class SimpleSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 16
    FIELDS = {
        1: ('inputs', TxInputType, p.FLAG_REPEATED),
        2: ('outputs', TxOutputType, p.FLAG_REPEATED),
        3: ('transactions', TransactionType, p.FLAG_REPEATED),
        4: ('coin_name', p.UnicodeType, 0),  # default='Bitcoin'
        5: ('version', p.UVarintType, 0),  # default=1
        6: ('lock_time', p.UVarintType, 0),  # default=0
        7: ('expiry', p.UVarintType, 0),
        8: ('overwintered', p.BoolType, 0),
    }

    def __init__(
        self,
        inputs: List[TxInputType] = None,
        outputs: List[TxOutputType] = None,
        transactions: List[TransactionType] = None,
        coin_name: str = None,
        version: int = None,
        lock_time: int = None,
        expiry: int = None,
        overwintered: bool = None
    ) -> None:
        self.inputs = inputs if inputs is not None else []
        self.outputs = outputs if outputs is not None else []
        self.transactions = transactions if transactions is not None else []
        self.coin_name = coin_name
        self.version = version
        self.lock_time = lock_time
        self.expiry = expiry
        self.overwintered = overwintered



# Automatically generated by pb2py
from .. import protobuf as p


class TxOutputBinType(p.MessageType):
    FIELDS = {
        1: ('amount', p.UVarintType, 0),  # required
        2: ('script_pubkey', p.BytesType, 0),  # required
        3: ('decred_script_version', p.UVarintType, 0),
    }

    def __init__(
        self,
        amount: int = None,
        script_pubkey: bytes = None,
        decred_script_version: int = None
    ) -> None:
        self.amount = amount
        self.script_pubkey = script_pubkey
        self.decred_script_version = decred_script_version



# Automatically generated by pb2py
from .. import protobuf as p


class DecryptedMessage(p.MessageType):
    MESSAGE_WIRE_TYPE = 52
    FIELDS = {
        1: ('message', p.BytesType, 0),
        2: ('address', p.UnicodeType, 0),
    }

    def __init__(
        self,
        message: bytes = None,
        address: str = None
    ) -> None:
        self.message = message
        self.address = address



#
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import time
import os
import atexit
import usb1
import sys

from ..protocol_v1 import ProtocolV1
from ..protocol_v2 import ProtocolV2
from . import Transport, TransportException

DEV_SafeWISE1 = (0x1209, 0xabba)
DEV_SafeWISE2 = (0xffff, 0x53c1)
DEV_SafeWISE2_BL = (0xffff, 0x53c0)

INTERFACE = 0
ENDPOINT = 1
DEBUG_INTERFACE = 1
DEBUG_ENDPOINT = 2


class WebUsbHandle:

    def __init__(self, device):
        self.device = device
        self.count = 0
        self.handle = None

    def open(self, interface):
        if self.count == 0:
            self.handle = self.device.open()
            if self.handle is None:
                if sys.platform.startswith('linux'):
                    args = ('Do you have udev rules installed? https://github.com/trezor/trezor-common/blob/master/udev/51-trezor.rules', )
                else:
                    args = ()
                raise IOError('Cannot open device', *args)
            self.handle.claimInterface(interface)
            self.count += 1

    def close(self, interface):
        if self.count == 1:
            self.handle.releaseInterface(interface)
            self.handle.close()
        if self.count > 0:
            self.count -= 1


class WebUsbTransport(Transport):
    '''
    WebUsbTransport implements transport over WebUSB interface.
    '''

    PATH_PREFIX = 'webusb'
    context = None

    def __init__(self, device, protocol=None, handle=None, debug=False):
        super(WebUsbTransport, self).__init__()

        if handle is None:
            handle = WebUsbHandle(device)

        if protocol is None:
            # force_v1 = os.environ.get('SafeWISE_TRANSPORT_V1', '0')
            force_v1 = True

            if is_safewise2(device) and not int(force_v1):
                protocol = ProtocolV2()
            else:
                protocol = ProtocolV1()

        self.device = device
        self.protocol = protocol
        self.handle = handle
        self.debug = debug

    def get_path(self):
        return "%s:%s" % (self.PATH_PREFIX, dev_to_str(self.device))

    @classmethod
    def enumerate(cls):
        if cls.context is None:
            cls.context = usb1.USBContext()
            cls.context.open()
            atexit.register(cls.context.close)
        devices = []
        for dev in cls.context.getDeviceIterator(skip_on_error=True):
            if not (is_safewise1(dev) or is_safewise2(dev) or is_safewise2_bl(dev)):
                continue
            if not is_vendor_class(dev):
                continue
            try:
                # workaround for issue #223:
                # on certain combinations of Windows USB drivers and libusb versions,
                # SafeWISE is returned twice (possibly because Windows know it as both
                # a HID and a WebUSB device), and one of the returned devices is
                # non-functional.
                dev.getProduct()
                devices.append(WebUsbTransport(dev))
            except usb1.USBErrorNotSupported:
                pass
        return devices

    def find_debug(self):
        if isinstance(self.protocol, ProtocolV2):
            # TODO test this
            # For v2 protocol, lets use the same WebUSB interface, but with a different session
            protocol = ProtocolV2()
            debug = WebUsbTransport(self.device, protocol, self.handle)
            return debug
        if isinstance(self.protocol, ProtocolV1):
            # For v1 protocol, find debug USB interface for the same serial number
            protocol = ProtocolV1()
            debug = WebUsbTransport(self.device, protocol, None, True)
            return debug
        raise TransportException('Debug WebUSB device not found')

    def open(self):
        interface = DEBUG_INTERFACE if self.debug else INTERFACE
        self.handle.open(interface)
        self.protocol.session_begin(self)

    def close(self):
        interface = DEBUG_INTERFACE if self.debug else INTERFACE
        self.protocol.session_end(self)
        self.handle.close(interface)

    def read(self):
        return self.protocol.read(self)

    def write(self, msg):
        return self.protocol.write(self, msg)

    def write_chunk(self, chunk):
        endpoint = DEBUG_ENDPOINT if self.debug else ENDPOINT
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        self.handle.handle.interruptWrite(endpoint, chunk)

    def read_chunk(self):
        endpoint = DEBUG_ENDPOINT if self.debug else ENDPOINT
        endpoint = 0x80 | endpoint
        while True:
            chunk = self.handle.handle.interruptRead(endpoint, 64)
            if chunk:
                break
            else:
                time.sleep(0.001)
        if len(chunk) != 64:
            raise TransportException('Unexpected chunk size: %d' % len(chunk))
        return bytearray(chunk)


def is_safewise1(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_SafeWISE1


def is_safewise2(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_SafeWISE2


def is_safewise2_bl(dev):
    return (dev.getVendorID(), dev.getProductID()) == DEV_SafeWISE2_BL


def is_vendor_class(dev):
    configurationId = 0
    altSettingId = 0
    return dev[configurationId][INTERFACE][altSettingId].getClass() == usb1.libusb1.LIBUSB_CLASS_VENDOR_SPEC


def dev_to_str(dev):
    return ':'.join(str(x) for x in ['%03i' % (dev.getBusNumber(), )] + dev.getPortNumberList())


TRANSPORT = WebUsbTransport



# Automatically generated by pb2py
from .. import protobuf as p


class NEMMosaicSupplyChange(p.MessageType):
    FIELDS = {
        1: ('namespace', p.UnicodeType, 0),
        2: ('mosaic', p.UnicodeType, 0),
        3: ('type', p.UVarintType, 0),
        4: ('delta', p.UVarintType, 0),
    }

    def __init__(
        self,
        namespace: str = None,
        mosaic: str = None,
        type: int = None,
        delta: int = None
    ) -> None:
        self.namespace = namespace
        self.mosaic = mosaic
        self.type = type
        self.delta = delta



#!/usr/bin/python
import subprocess
import os
import json
import time
import ecdsa
import hashlib
import binascii
from google.protobuf.descriptor_pb2 import FileDescriptorSet

PROTOBUF_PROTO_DIR=os.environ.get('PROTOBUF_PROTO_DIR', '/usr/include/')
SafeWISE_PROTO_DIR=os.environ.get('SafeWISE_PROTO_DIR', '../protob/')

def compile_config():
    cmd = "protoc --python_out=../signer/ -I" + PROTOBUF_PROTO_DIR + " -I./ config.proto"
    subprocess.check_call(cmd.split(), cwd=SafeWISE_PROTO_DIR)

def parse_json():
    return json.loads(open('config.json', 'r').read())


def get_compiled_proto():
    # Compile safewise.proto to binary format
    pdir = os.path.abspath(SafeWISE_PROTO_DIR)
    pfile = os.path.join(pdir, "messages.proto")
    cmd = "protoc --include_imports -I" + PROTOBUF_PROTO_DIR + " -I" + pdir  + " " + pfile + " -osafewise.bin"

    subprocess.check_call(cmd.split())

    # Load compiled protocol description to string
    proto = open('safewise.bin', 'r').read()
    os.unlink('safewise.bin')

    # Parse it into FileDescriptorSet structure
    compiled = FileDescriptorSet()
    compiled.ParseFromString(proto)
    return compiled

def compose_message(json, proto):
    import config_pb2

    cfg = config_pb2.Configuration()
    cfg.valid_until = 2147483647 # maxint
    cfg.wire_protocol.MergeFrom(proto)

    for url in json['whitelist_urls']:
        cfg.whitelist_urls.append(str(url))

    for url in json['blacklist_urls']:
        cfg.blacklist_urls.append(str(url))

    for dev in json['known_devices']:
        desc = cfg.known_devices.add()
        desc.vendor_id = int(dev[0], 16)
        desc.product_id = int(dev[1], 16)

    return cfg.SerializeToString()

def sign_message(data, key):
    if key.startswith('-----BEGIN'):
        key = ecdsa.keys.SigningKey.from_pem(key)
    else:
        key = ecdsa.keys.SigningKey.from_secret_exponent(secexp = int(key, 16), curve=ecdsa.curves.SECP256k1, hashfunc=hashlib.sha256)

    verify = key.get_verifying_key()
    print "Verifying key:"
    print verify.to_pem()

    return key.sign_deterministic(data, hashfunc=hashlib.sha256)

def pack_datafile(filename, signature, data):
    if len(signature) != 64:
        raise Exception("Signature must be 64 bytes long")

    fp = open(filename, 'w')
    fp.write(binascii.hexlify(signature))
    fp.write(binascii.hexlify(data))
    fp.close()

    print "Signature and data stored to", filename

if __name__ == '__main__':
    key = ''
    print "Paste ECDSA private key (in PEM format or SECEXP format) and press Enter:"
    while True:
        inp = raw_input()
        if inp == '':
            break

        key += inp + "\n"

    # key = open('sample.key', 'r').read()

    compile_config()
    json = parse_json()
    proto = get_compiled_proto()

    data = compose_message(json, proto)
    signature = sign_message(data, key)

    pack_datafile('config_signed.bin', signature, data)
